"""
The `VSPEC` star model is designed modularly to allow
for both simple and complex behaviors. Currently, it
is represented by a rectangular grid of points on the stellar
surface, each assigned an effective temperature. At any given
time, the model computes the surface coverage fractions of
each temperature visible to the observer, accounting for the
spherical geometry, limb darkening, and any occultation
by a transiting planet.

Once the surface coverage is calculated, the fractions of each component are
passed allong to VSPEC to compute the spectrum.

The attributes of the ``Star`` class describe the bulk properties of the star,
including radius, period, and the effective temperature of quiet photosphere.
Herein we refer to this temperature as the photosphere temperature to differentiate
it from the temperature of spots, faculae, or other sources of variability.
"""
from typing import Tuple, Union
import warnings
import numpy as np
from astropy import units as u
from astropy.units.quantity import Quantity

from vspec_vsm.coordinate_grid import CoordinateGrid
from vspec_vsm.helpers import (
    get_angle_between, proj_ortho,
    calc_circ_fraction_inside_unit_circle,
    clip_teff
)
from vspec_vsm.spots import SpotCollection, SpotGenerator
from vspec_vsm.faculae import FaculaCollection, FaculaGenerator, Facula
from vspec_vsm.flares import FlareCollection, FlareGenerator
from vspec_vsm.granules import Granulation

class InsufficientResolutionWarning(Warning):
    """Transit calculated but there is not
    enough grid resolution to do it accurately.
    """

class Star:
    """
    Star object representing a variable star.

    Parameters
    ----------
    Teff : astropy.units.Quantity
        Effective temperature of the stellar photosphere.
    radius : astropy.units.Quantity
        Stellar radius.
    period : astropy.units.Quantity
        Stellar rotational period.
    spots : SpotCollection
        Initial spots on the stellar surface.
    faculae : FaculaCollection
        Initial faculae on the stellar surface.
    distance : astropy.units.Quantity , default=1*u.pc
        Distance to the star.
    Nlat : int, default=500
        The number of latitude points on the stellar sufrace.
    Nlon : int, default=1000
        The number of longitude points on the stellar surface.
    gridmaker : CoordinateGrid, default=None
        A `CoordinateGrid` object to create the stellar sufrace grid.
    flare_generator : FlareGenerator, default=None
        Flare generator object.
    spot_generator : SpotGenerator, default=None
        Spot generator object.
    fac_generator : FaculaGenerator, default=None
        Facula generator object.
    ld_params : list, default=[0, 1, 0]
        Limb-darkening parameters.

    Attributes
    ----------
    Teff : astropy.units.Quantity
        Effective temperature of the stellar photosphere.
    radius : astropy.units.Quantity
        Stellar radius.
    distance : astropy.units.Quantity
        Distance to the star.
    period : astropy.units.Quantity
        Stellar rotational period.
    spots : SpotCollection
        Spots on the stellar surface.
    faculae : FaculaCollection
        Faculae on the stellar surface.
    gridmaker : CoordinateGrid
        Object to create the coordinate grid of the surface.
    map : astropy.units.Quantity
        Pixel map of the stellar surface.
    flare_generator : FlareGenerator
        Flare generator object.
    spot_generator : SpotGenerator
        Spot generator object.
    fac_generator : FaculaGenerator
        Facula generator object.
    u1 : float
        Limb-darkening parameter u1.
    u2 : float
        Limb-darkening parameter u2.
    """

    def __init__(self, teff: u.Quantity,
                 radius: u.Quantity,
                 period: u.Quantity,
                 spots: SpotCollection,
                 faculae: FaculaCollection,
                 grid_params: Union[int,Tuple[int, int]] = 1000,
                 flare_generator: FlareGenerator = None,
                 spot_generator: SpotGenerator = None,
                 fac_generator: FaculaGenerator = None,
                 granulation: Granulation = None,
                 u1: float = 0,
                 u2: float = 0,
                 rng: np.random.Generator = np.random.default_rng()
                 ):
        self.teff = teff
        self.radius = radius
        self.period = period
        self.spots:SpotCollection = spots
        self.faculae:FaculaCollection = faculae
        self.rng = rng
        self.gridmaker = CoordinateGrid.new(grid_params)
        self.grid_params = grid_params
        
        self.faculae.set_gridmaker(self.gridmaker)
        self.spots.set_gridmaker(self.gridmaker)

        if flare_generator is None:
            self.flare_generator = FlareGenerator.off(
                rng=self.rng
            )
        else:
            self.flare_generator = flare_generator
        self.flares = None

        if spot_generator is None:
            self.spot_generator = SpotGenerator.off(
                grid_params=grid_params,
                gridmaker=self.gridmaker,
                rng=self.rng
            )
        else:
            self.spot_generator = spot_generator
            try:
                if self.spot_generator.gridmaker != self.gridmaker:
                    self.spot_generator.gridmaker = self.gridmaker
            except TypeError:
                self.spot_generator.gridmaker = self.gridmaker
            

        if fac_generator is None:
            self.fac_generator = FaculaGenerator.off(
                grid_params=grid_params,
                gridmaker=self.gridmaker,
                rng=self.rng
            )
        else:
            self.fac_generator = fac_generator
            try:
                if self.fac_generator.gridmaker != self.gridmaker:
                    self.fac_generator.gridmaker = self.gridmaker
            except TypeError:
                self.fac_generator.gridmaker = self.gridmaker
        if granulation is None:
            self.granulation = Granulation.off(seed=0)
        else:
            self.granulation = granulation
        self.u1 = u1
        self.u2 = u2
        self.set_spot_grid()
        self.set_fac_grid()

    def set_spot_grid(self):
        """
        Set the gridmaker for each spot to be the same.
        """
        for spot in self.spots.spots:
            spot.gridmaker = self.gridmaker

    def set_fac_grid(self):
        """
        Set the gridmaker for each facula to be the same.
        """
        for fac in self.faculae.faculae:
            fac.gridmaker = self.gridmaker

    @property
    def map(self):
        """
        Create a map of the stellar surface based on spots.

        Returns
        -------
        pixelmap : astropy.units.Quantity , Shape(self.gridmaker.Nlon,self.gridmaker.Nlat)
            Map of stellar surface with effective temperature assigned to each pixel.
        """
        return self.spots.map_pixels(self.radius, self.teff)

    def age(self, time):
        """
        Age the spots and faculae on the stellar surface according
        to their own `age` methods. Remove the spots that have decayed.

        Parameters
        ----------
        time : astropy.units.Quantity 
            Length of time to age the features on the stellar surface.
            For most realistic behavior, `time` should be much less than
            spot or faculae lifetime.
        """
        self.spots.age(time)
        self.faculae.age(time)

    def add_spot(self, spot):
        """
        Add one or more spots to the stellar surface.

        Parameters
        ----------
        spot : StarSpot or sequence of StarSpot
            The `StarSpot` object(s) to add.
        """
        self.spots.add_spot(spot)

    def add_fac(self, facula):
        """
        Add one or more faculae to the stellar surface.

        Parameters
        ----------
        facula : Facula or sequence of Facula
            The Facula object(s) to add.

        """
        self.faculae.add_faculae(facula)

    def get_mu(self, lat0: u.Quantity, lon0: u.Quantity) -> np.ndarray:
        """
        Get the cosine of the angle from disk center.

        Parameters
        ----------
        lat0 : astropy.units.Quantity
            The sub-observer latitude.
        lon0 : astropy.units.Quantity
            The sub-observer longitude

        Returns
        -------
        mu : np.ndarray
            An array of cos(x) where x is
            the angle from disk center.

        """
        return self.gridmaker.cos_angle_from_disk_center(lat0, lon0)

    def ld_mask(self, mu) -> np.ndarray:
        """
        Get a translucent mask based on limb darkeining parameters.

        Parameters
        ----------
        mu : np.ndarray
            The cosine of the angle from disk center

        Returns
        -------
        mask : np.ndarray
            The limb-darkened mask.

        Notes
        -----
        To account for apparent size effect of points on the
        stellar surface, we add a factor of 1 to `u1`. This way,
        in the (Lambertian) case of no-limb darkening, the user
        can set ``u1 = 0``, ``u2 = 0``
        """
        mask = 1 - (self.u1+1) * (1 - mu) - self.u2 * (1 - mu)**2
        behind_star = mu < 0.
        mask[behind_star] = 0
        return mask

    def ld_mask_for_plotting(self, mu) -> np.ndarray:
        """
        Same as above, but does not add the extra 1
        because that is already accounted for by the projection.
        """
        mask = 1 - (self.u1) * (1 - mu) - self.u2 * (1 - mu)**2
        behind_star = mu < 0.
        mask[behind_star] = 0
        return mask

    def get_jacobian(self) -> np.ndarray:
        """
        Get the relative area of each point.

        Returns
        -------
        jacobian : np.ndarray
            The area of each point
        """
        return self.gridmaker.area

    def add_faculae_to_map(
        self,
        lat0: u.Quantity,
        lon0: u.Quantity
    ):
        """
        Add the faculae to the surface map.

        Parameters
        ----------
        lat0 : astropy.units.Quantity
            The sub-observer latitude.
        lon0 : astropy.units.Quantity
            The sub-observer longitude.

        Returns
        -------
        teffmap : astropy.units.Quantity
            A temperature map of the surface
        """
        map_from_spots = self.map
        mu = self.get_mu(lat0, lon0)
        faculae: Tuple[Facula] = self.faculae.faculae
        for facula in faculae:
            angle = get_angle_between(lat0, lon0, facula.lat, facula.lon)
            inside_fac = facula.map_pixels(self.radius)
            if not np.any(inside_fac):  # the facula is too small
                pass
            else:
                fracs = facula.fractional_effective_area(angle)
                dteff_wall, dteff_floor = fracs.keys()
                frac = fracs[dteff_wall].value
                mu_of_fac_pix = mu[inside_fac]
                border_mu = np.percentile(mu_of_fac_pix, 100*frac)
                wall_pix = inside_fac & (mu <= border_mu)
                floor_pix = inside_fac & (mu > border_mu)
                teff_wall = clip_teff(dteff_wall + self.teff)
                teff_floor = clip_teff(dteff_floor + self.teff)
                map_from_spots[wall_pix] = teff_wall
                map_from_spots[floor_pix] = teff_floor
        return map_from_spots

    def get_pl_frac(
        self,
        angle_past_midtransit: u.Quantity,
        orbit_radius: u.Quantity,
        planet_radius: u.Quantity,
        inclination: u.Quantity
    ):
        """
        Get planet fraction

        Parameters
        ----------
        angle_past_midtransit : astropy.units.Quantity
            The phase of the planet past the 180 degree mid transit point.
        orbit_radius : astropy.units.Quantity
            The radius of the planet's orbit.
        radius : astropy.units.Quantity
            The radius of the planet.

        inclination : astropy.units.Quantity
            The inclination of the planet. 90 degrees is transiting.
        """
        x = (orbit_radius/self.radius * np.sin(angle_past_midtransit)
             ).to_value(u.dimensionless_unscaled)
        y = (orbit_radius/self.radius * np.cos(angle_past_midtransit)
             * np.cos(inclination)).to_value(u.dimensionless_unscaled)
        rad = (planet_radius/self.radius).to_value(u.dimensionless_unscaled)
        return 1-calc_circ_fraction_inside_unit_circle(x, y, rad)

    def get_transit_mask(
        self,
        lat0: u.Quantity,
        lon0: u.Quantity,
        orbit_radius: u.Quantity,
        radius: u.Quantity,
        phase: u.Quantity,
        inclination: u.Quantity
    ):
        """
        Get a mask describing which pixels are covered by a transiting planet.

        Parameters
        ----------
        lat0 : astropy.units.Quantity
            The sub-observer latitude.
        lon0 : astropy.units.Quantity
            The sub-observer longitude.
        orbit_radius : astropy.units.Quantity
            The radius of the planet's orbit.
        radius : astropy.units.Quantity
            The radius of the planet.
        phase : astropy.units.Quantity
            The phase of the planet. 180 degrees is mid transit.
        inclination : astropy.units.Quantity
            The inclination of the planet. 90 degrees is transiting.

        Returns
        -------
        mask : np.ndarray
            The fraction of each pixel that is covered by the planet.
        pl_frac : float
            The fraction of the planet that is visible to the observer.
        """
        eclipse = False
        if np.cos(phase) > 0:
            eclipse = True
        angle_past_midtransit: u.Quantity = phase - 180*u.deg
        x: float = (orbit_radius/self.radius * np.sin(angle_past_midtransit)
             ).to_value(u.dimensionless_unscaled)
        y: float = (orbit_radius/self.radius * np.cos(angle_past_midtransit)
             * np.cos(inclination)).to_value(u.dimensionless_unscaled)
        rp_rs: float = (radius/self.radius).to_value(u.dimensionless_unscaled)
        if np.sqrt(x**2 + y**2) > 1 + 2*rp_rs:  # no transit
            return self.gridmaker.zeros().astype('bool'), 1.0
        elif eclipse:
            planet_fraction = self.get_pl_frac(
                angle_past_midtransit, orbit_radius, radius, inclination)
            return self.gridmaker.zeros().astype('bool'), planet_fraction
        else:
            llat, llon = self.gridmaker.grid()
            xcoord, ycoord = proj_ortho(lat0, lon0, llat, llon)
            mu = self.gridmaker.cos_angle_from_disk_center(lat0, lon0)
            area = self.gridmaker.area  # area in units of 4pi steradians ie adds to 1
            point_radii = 2*np.sqrt(area)  # radius of each pixel in radians. The 2
                                           # comes from the 4 in 4pi steradians
            proj_radii = point_radii*mu  # radius of each pixel in projected coords
            # distances in projected coords
            rad_map = np.sqrt((xcoord-x)**2 + (ycoord-y)**2)
            # case 1: Point is completely outside transit radius
            case1 = (rad_map > rp_rs + 2*proj_radii) | np.isnan(rad_map)

            covered_value = np.where(~case1, 1, 0).astype('float')
            if np.any(np.isnan(covered_value)):
                raise ValueError('NaN in covered_value')
            indicies = np.argwhere(~case1)
            if len(indicies)>0:
                relevent_radii = proj_radii[~case1]
                rad_mean = np.mean(relevent_radii)
                if rad_mean > 0.5*rp_rs:
                    area_mean = np.pi*rad_mean**2
                    area_pl = np.pi*rp_rs**2
                    target_area = area_pl/4
                    factor = area_mean/target_area
                    
                    warnings.warn(
                        f'Pixel resolution too low. Increase by factor of {factor:.2f}',
                        InsufficientResolutionWarning
                    )
                    
            for index in indicies:
                s = index if index is int else tuple(index)
                dist_from_transit_center: float = rad_map[s]
                # gauss_sigma = proj_radii[s]
                sigma_x = point_radii[s]*mu[s]
                sigma_y = point_radii[s]
                if rp_rs > dist_from_transit_center+2*sigma_x:
                    covered_value[s] = 1.
                else:
                    x = np.linspace(-3*sigma_x, 3*sigma_x, 100)
                    xx, yy = np.meshgrid(x, x)
                    # zz = 1/(2*np.pi*gauss_sigma**2)*np.exp(-(xx**2 + yy**2)
                    #                                        / (2*gauss_sigma**2))
                    zz = 1/(2*np.pi*sigma_x*sigma_y)*np.exp(-0.5*(xx/sigma_x)**2 - 0.5*(yy/sigma_y)**2)
                    dist = np.sqrt((xx-dist_from_transit_center)**2 + (yy)**2)
                    overlap = np.where(dist < rp_rs, zz, 0)
                    overlap = np.trapz(overlap, x, axis=1)
                    overlap = np.trapz(overlap, x, axis=0)
                    covered_value[s] = overlap
            if np.any(np.isnan(covered_value)):
                raise ValueError('NaN in covered_value')
            return covered_value, 1.0

    def calc_coverage(
        self,
        sub_obs_coords: dict,
        granulation_fraction: float = 0.0,
        orbit_radius: u.Quantity = 1*u.AU,
        planet_radius: u.Quantity = 1*u.R_earth,
        phase: u.Quantity = 90*u.deg,
        inclination: u.Quantity = 0*u.deg

    ):
        """
        Calculate coverage

        Calculate coverage fractions of various Teffs on stellar surface
        given coordinates of the sub-observation point.

        Parameters
        ----------
        sub_obs_coord : dict
            A dictionary giving coordinates of the sub-observation point.
            This is the point that is at the center of the stellar disk from the view of
            an observer. Format: {'lat':lat,'lon':lon} where lat and lon are
            `astropy.units.Quantity` objects.
        granulation_fraction : float
            The fraction of the quiet photosphere that has a lower Teff due to granulation

        Returns
        -------
        total_data : dict
            Dictionary with Keys as Teff quantities and Values as surface fraction floats.
        covered_data : dict
            Dictionary with Keys as Teff quantities and Values as surface fraction floats covered
            by a transiting planet.
        pl_frac : float
            The fraction of the planet that is visble. This is in case of an eclipse.
        """
        cos_c = self.get_mu(sub_obs_coords['lat'], sub_obs_coords['lon'])
        ld = self.ld_mask(cos_c)
        jacobian = self.get_jacobian()

        surface_map = self.add_faculae_to_map(
            sub_obs_coords['lat'], sub_obs_coords['lon'])
        covered, pl_frac = self.get_transit_mask(
            sub_obs_coords['lat'], sub_obs_coords['lon'],
            orbit_radius=orbit_radius,
            radius=planet_radius,
            phase=phase,
            inclination=inclination
        )

        teffs = np.unique(surface_map)
        total_data = {}
        covered_data = {}
        total_area = np.sum(ld*jacobian)
        for teff in teffs:
            pix_has_teff = np.where(surface_map == teff, 1, 0)
            nominal_area = np.sum(pix_has_teff*ld*jacobian)
            covered_area = np.sum(pix_has_teff*ld*jacobian*(covered))
            total_data[teff] = nominal_area/total_area
            covered_data[teff] = covered_area/total_area
        granulation_teff = self.teff - self.granulation.dteff
        # initialize. This way it's okay if there's something else with that Teff too.
        if granulation_teff not in teffs:
            total_data[granulation_teff] = 0
            covered_data[granulation_teff] = 0

        phot_frac = total_data[self.teff]
        total_data[self.teff] = phot_frac * (1-granulation_fraction)
        total_data[granulation_teff] += phot_frac * granulation_fraction
        phot_frac = covered_data[self.teff]
        covered_data[self.teff] = phot_frac * (1-granulation_fraction)
        covered_data[granulation_teff] += phot_frac * granulation_fraction

        return total_data, covered_data, pl_frac

    def birth_spots(self, time):
        """
        Create new spots from a spot generator.

        Parameters
        ----------
        time : astropy.units.Quantity 
            Time over which these spots should be created.

        """
        self.spots.add_spot(self.spot_generator.birth_spots(time, self.radius))

    def birth_faculae(self, time):
        """
        Create new faculae from a facula generator.

        Parameters
        ----------
        time : astropy.units.Quantity 
            Time over which these faculae should be created.


        """
        self.faculae.add_faculae(
            self.fac_generator.birth_faculae(time, self.radius))

    def average_teff(self, sub_obs_coords):
        """
        Calculate the average Teff of the star given a sub-observation point
        using the Stephan-Boltzman law. This can approximate a lightcurve for testing.

        Parameters
        ----------
        sub_obs_coords : dict
            A dictionary containing coordinates of the sub-observation point. This is the 
            point that is at the center of the stellar disk from the view of an observer. 
            Format: {'lat':lat,'lon':lon} where lat and lon are `astropy.units.Quantity` objects.

        Returns
        -------
        astropy.units.Quantity 
            Bolometric average Teff of stellar disk.

        """
        dat, _, _ = self.calc_coverage(sub_obs_coords)
        num = 0
        den = 0
        for teff, value in dat.items():
            num += teff**4 * value
            den += value
        return ((num/den)**(0.25)).to(u.K)

    def plot_surface(
        self,
        lat0: u.Quantity,
        lon0: u.Quantity,
        ax = None,
        orbit_radius: u.Quantity = 1*u.AU,
        radius: u.Quantity = 1*u.R_earth,
        phase: u.Quantity = 90*u.deg,
        inclination: u.Quantity = 0*u.deg,
        nlon: int = 1000,
        nlat: int = 500
    ):
        """
        Add the transit to the surface map and plot.
        """
        import matplotlib.pyplot as plt
        import cartopy.crs as ccrs
        from cartopy.mpl.geoaxes import GeoAxes
        ax: GeoAxes
        proj = ccrs.Orthographic(
            central_latitude=lat0.to_value(u.deg),
            central_longitude=lon0.to_value(u.deg)
        )
        if ax is None:
            _, ax = plt.subplots(subplot_kw={'projection': proj})
            ax: GeoAxes = ax
        elif ax.projection != proj:
            ax.projection = (proj)
        covered, pl_frac = self.get_transit_mask(
            lat0, lon0,
            orbit_radius=orbit_radius,
            radius=radius,
            phase=phase,
            inclination=inclination
        )
        map_with_faculae = self.add_faculae_to_map(lat0, lon0)

        lats, lons, data = self.gridmaker.display_grid(
            nlat, nlon, map_with_faculae)

        lats = (lats*u.rad).to_value(u.deg)
        lons = (lons*u.rad).to_value(u.deg)
        data = data.to_value(u.K)
        im = ax.pcolormesh(lons, lats, data.T,
                           transform=ccrs.PlateCarree())
        plt.colorbar(im, ax=ax, label=r'$T_{\rm eff}$ (K)')

        _, _, data = self.gridmaker.display_grid(nlat, nlon, covered)

        transit_mask = np.where(data > 0.5, 1, np.nan)
        zorder = 100 if pl_frac == 1. else -100
        ax.contourf(lons, lats, transit_mask.T, colors='k', alpha=1,
                    transform=ccrs.PlateCarree(), zorder=zorder)
        mu = self.get_mu(lat0, lon0)
        ld = self.ld_mask_for_plotting(mu)
        alpha = 1-ld.T/np.max(ld.T)
        ax.imshow(np.ones_like(ld), extent=(lons.min(), lons.max(), lats.min(), lats.max()),
                  transform=ccrs.PlateCarree(), origin='lower', alpha=alpha, cmap=plt.cm.get_cmap('gray'), zorder=100)

    def get_flares_over_observation(self, time_duration: Quantity):
        """
        Generate a collection of flares over a specified observation period.

        Parameters
        ----------
        time_duration: astropy.units.Quantity 
            The duration of the observation period.

        Notes
        -----
        This method uses the `FlareGenerator` attribute of the `Star` object to generate
        a distribution of flare energies, and then generates a series of flares over the
        specified observation period using these energies. The resulting collection of
        flares is stored in the `Star`'s `flares` attribute.
        """
        flares = self.flare_generator.generate_flare_series(time_duration)
        self.flares = FlareCollection(flares)

    def get_flare_int_over_timeperiod(self, tstart: Quantity, tfinish: Quantity, sub_obs_coords):
        """
        Compute the total flare integral over a specified time period and sub-observer point.

        Parameters
        ----------
        tstart: astropy.units.Quantity 
            The start time of the period.
        tfinish: astropy.units.Quantity 
            The end time of the period.
        sub_obs_coords : dict
            A dictionary containing coordinates of the sub-observation point. This is the 
            point that is at the center of the stellar disk from the view of an observer. 
            Format: {'lat':lat,'lon':lon} where lat and lon are `astropy.units.Quantity` objects.


        Returns
        -------
        flare_timeareas: list of dict
            List of dictionaries containing flare temperatures and integrated
            time-areas. In the format [{'Teff':9000*u.K,'timearea'=3000*u.Unit('km2 hr)},...]

        Notes
        -----
        This method computes the total flare integral over each flare in the `flares` 
        attribute of the `Star` object that falls within the specified time period and is visible
        from the sub-observer point defined by `sub_obs_coords`. The result is returned
        as a list of dictionaries representing the teff and total flare integral of each flare.
        """
        flare_timeareas = self.flares.get_flare_integral_in_timeperiod(
            tstart, tfinish, sub_obs_coords)
        return flare_timeareas

    def generate_mature_spots(self, coverage: float):
        """
        Generate new mature spots with a specified coverage.

        Parameters
        ----------
        coverage: float
            The coverage of the new spots.

        Notes
        -----
        This method uses the `SpotGenerator` attribute of the current object to generate a
        set of new mature spots with the specified coverage. The new spots are added to 
        the object's `spots` attribute and the pixel map is updated using the new spots.
        """
        new_spots = self.spot_generator.generate_mature_spots(
            coverage, self.radius)
        self.spots.add_spot(new_spots)

    def get_granulation_coverage(self, time: u.Quantity) -> np.ndarray:
        """
        Calculate the coverage by granulation at each point in `time`.

        Parameters
        ----------
        time : astropy.units.Quantity
            The points on the time axis.

        Returns
        -------
        np.ndarray
            The coverage corresponding to each `time` point.
        """
        if self.granulation is None:
            return np.zeros(shape=time.shape)
        else:
            coverage = self.granulation.get_coverage(time)
            return np.where(np.isnan(coverage), 0, coverage)



"""
Miscalaneous helper functions
"""

import astropy.units as u
import numpy as np
import warnings

from vspec_vsm import config

def round_teff(teff):
    """
    Round the effective temperature to the nearest integer.
    The goal is to reduce the number of unique effective temperatures
    while not affecting the accuracy of the model.

    Parameters
    ----------
    teff : astropy.units.Quantity
        The temperature to round.

    Returns
    -------
    astropy.units.Quantity
        The rounded temperature.

    Notes
    -----
    This function rounds the given effective temperature to the nearest integer value. It is designed to decrease the number of unique effective temperatures while maintaining the accuracy of the model.

    Examples
    --------
    >>> teff = 1234.56 * u.K
    >>> rounded_teff = round_teff(teff)
    >>> print(rounded_teff)
    1235 K

    >>> teff = 2000.4 * u.K
    >>> rounded_teff = round_teff(teff)
    >>> print(rounded_teff)
    2000 K

    """

    val = teff.value
    unit = teff.unit
    return int(round(val)) * unit

def get_angle_between(
    lat1: u.Quantity,
    lon1: u.Quantity,
    lat2: u.Quantity,
    lon2: u.Quantity
):
    """
    Compute the angle between two coordintates in
    lat/lon space.

    Parameters
    ----------
    lat1 : astropy.units.Quantity
        The latitude of the first coordinate
    lon1 : astropy.units.Quantity
        The longitude of the first coordinate
    lat2 : astropy.units.Quantity
        The latitude of the second coordinate
    lon2 : astropy.units.Quantity
        The longitude of the second coordinate

    Returns
    -------
    alpha : astropy.units.Quantity
        The angle between the coordinates.

    Notes
    -----
    This function computes the angle between two coordinates specified in latitude and longitude space using the spherical law of cosines:

    .. math::
        \\mu = \\sin{\\phi_1} \\sin{\\phi_2} + \\cos{\\phi_1} \\cos{\\phi_2} \\cos{(\\lambda_1 - \\lambda_2)}

    where :math:`(\\phi_1, \\lambda_1)` and :math:`(\\phi_2, \\lambda_2)` are the latitude and longitude pairs of the two coordinates, respectively. The angle between the coordinates is then computed as:

    .. math::
        \\alpha = \\arccos{\\mu}

    The computed angle is returned as an `astropy.units.Quantity` object.

    Examples
    --------
    >>> lat1 = 30 * u.deg
    >>> lon1 = 45 * u.deg
    >>> lat2 = 40 * u.deg
    >>> lon2 = 60 * u.deg
    >>> angle = get_angle_between(lat1, lon1, lat2, lon2)
    >>> print(angle)
    0.27581902582503454 rad

    """
    mu = (np.sin(lat1) * np.sin(lat2)
          + np.cos(lat1) * np.cos(lat2)
          * np.cos(lon1-lon2))
    return np.arccos(mu)

def proj_ortho(
    lat0: u.Quantity,
    lon0: u.Quantity,
    lats: u.Quantity,
    lons: u.Quantity
):
    """
    Use the spherical law of cosines to project each
    lat/lon point onto the x-y plane.

    Parameters
    ----------
    lat0 : astropy.units.quantity
        The latitude of the central point.
    lon0 : astropy.units.quantity
        The longitude of the central point.
    lats : astropy.units.quantity
        The latitude of the other points.
    lons : astropy.units.quantity
        The longitude of the other points.

    Returns
    -------
    x : np.ndarray
        The x coordinates of the points in the
        orthographic projection
    y : np.ndarray
        The y coordinates of the points in the
        orthographic projection

    Notes
    -----
    This function uses the spherical law of cosines to project each lat/lon point
    onto the x-y plane of an orthographic projection.

    Spherical law of cosines:

    .. math::

        \\cos{c} = \\cos{a} \\cos{b} + \\sin{a} \\sin{b} \\cos{C}

    We want to find :math:`C`.
        - :math:`b` is the colatitude of the central point.
        - :math:`c` is the colatitude of the other points.
        - :math:`a` is the distance of each point from the center (i.e. :math:`\\cos{\\mu}`)
    This can be difficult for special cases, but this code attempts
    to account for those.

    Examples
    --------
    >>> lat0 = 30 * u.deg
    >>> lon0 = 45 * u.deg
    >>> lats = [40, 50, 60] * u.deg
    >>> lons = [60, 70, 80] * u.deg
    >>> x, y = proj_ortho(lat0, lon0, lats, lons)
    >>> print(x)
    [0.19825408 0.27164737 0.28682206]
    >>> print(y)
    [0.18671294 0.37213692 0.54519419]

    """
    if not isinstance(lon0, u.Quantity):
        raise TypeError('`lon0` must be a Quantity')
    if not isinstance(lat0, u.Quantity):
        raise TypeError('`lat0` must be a Quantity')
    if not isinstance(lons, u.Quantity):
        raise TypeError('`lons` must be a Quantity')
    if not isinstance(lats, u.Quantity):
        raise TypeError('`lats` must be a Quantity')
    # Cast all to radians
    lon0: float = lon0.to_value(u.rad)
    lat0: float = lat0.to_value(u.rad)
    lons: np.ndarray = lons.to_value(u.rad)
    lats: np.ndarray = lats.to_value(u.rad)
    b = np.pi/2 - lat0
    c = np.pi/2 - lats
    if b == 0:  # centered on north pole
        a = c
        C = np.where(c <= np.pi/2, np.pi-lons, np.nan)
    elif b == np.pi:  # centered on south pole
        a = c
        C = np.where(c >= np.pi/2, lons, np.nan)
    else:  # centered elsewhere
        mu = (np.sin(lat0) * np.sin(lats)
              + np.cos(lat0) * np.cos(lats)
              * np.cos(lon0-lons))
        a = np.arccos(mu)
        num = np.cos(c) - np.cos(a)*np.cos(b)
        den = np.sin(a)*np.sin(b)
        den[den==0]=1e-20 # supress warning. This is overwritten anyway.
        cos_C = np.where(
            mu < 1,
            num / den,
            1
        )
        cos_C = np.round(cos_C, 4)
        C = np.where(
            lons == lon0,
            0, np.arccos(cos_C)
        )
        C = np.where(
            (lons == lon0) & (lats < lat0),
            np.pi, C
        )
        C = np.where(
            mu >= 0,
            C,
            np.nan
        )
        C = np.where(
            np.sin(lons-lon0) < 0,
            -C, C
        )
    x = np.sin(C)*np.sin(a)
    y = np.cos(C)*np.sin(a)
    return x, y


def calc_circ_fraction_inside_unit_circle(x, y, r):
    """
    Calculate what fraction of a circle sits inside
    the unit circle.

    This is useful for modeling secondary eclipse.

    Parameters
    ----------
    x : float
        The x coordinate of the circle's center.
    y : float
        The y coordinate of the circle's center.
    r : float
        The circle's radius.

    Returns
    -------
    float
        The fraction of the circle that lies inside the unit circle.

    Raises
    ------
    ValueError
        If the circle's radius exceeds 1.


    Notes
    -----
    The calculation is based on the circle-circle intersection math presented in [1].

    References
    ----------
    [1].  Weisstein, Eric W. "Circle-Circle Intersection."
        From MathWorld--A Wolfram Web Resource. https://mathworld.wolfram.com/Circle-CircleIntersection.html 
    """
    x0, y0, r0 = 0, 0, 1
    d = np.sqrt((x-x0)**2 + (y-y0)**2)
    if r > 1:
        raise ValueError('R must be less than 1.')

    if d > (r0 + r):
        return 0.0
    if d <= (r0 - r):
        return 1.0
    if d == 0 and r == 1:  # it is the unit circle
        return 1.0

    R = r0
    d1 = (d**2 - r**2 + R**2) / (2*d)
    d2 = (d**2 + r**2 - R**2) / (2*d)

    def area(_R, _d):
        return _R**2 * np.arccos(_d/_R) - _d * np.sqrt(_R**2 - _d**2)
    A1 = area(R, d1)
    A2 = area(r, d2)
    area_of_intersection = A1+A2

    area_of_circle = np.pi*r**2
    return area_of_intersection/area_of_circle

def clip_teff(teff: u.Quantity):
    """
    Clip an effective temperature value to ensure it is within the bounds of
    available models.

    Parameters
    ----------
    teff : astropy.units.Quantity
        The effecitve temperature to clip

    Returns
    -------
    astropy.units.Quantity
        The clipped effective temperature.
    """
    low, high = config.grid_teff_bounds
    if teff > high:
        warnings.warn(
            f'Teff of {teff:.1f} too high, clipped to {high:.1f}', RuntimeWarning)
        return high
    elif teff < low:
        warnings.warn(
            f'Teff of {teff:.1f} too low, clipped to {low:.1f}', RuntimeWarning)
        return low
    else:
        return teff



"""
Spots
-----

This contains all the code governing
the behavior of spots.

"""
from typing import List, Union, Tuple
import typing as Typing

import numpy as np
from astropy import units as u
from astropy.units.quantity import Quantity

from vspec_vsm.coordinate_grid import CoordinateGrid
from vspec_vsm.config import MSH


class StarSpot:
    """
    A Star Spot

    Our star spot model is nearly entirely based on
    observations of the sun. Sun spots can be resolved
    and are well-studied, whereas spots on other stars
    (e.g., M dwarfs) can only observed indirectly. We
    therefore designed our spot model to mimic sun spots
    but with parameterized values for spot temperature and 
    lifetime that can be matched to observations of other 
    stellar types and ages.

    Parameters
    ----------
    lat : astropy.units.Quantity 
        Latitude of spot center. North is positive.
    lon : astropy.units.Quantity 
        Longitude of spot center. East is positive.
    Amax : astropy.units.Quantity 
        The maximum area a spot reaches before it decays.
    A0 : astropy.units.Quantity 
        The current spot area.
    Teff_umbra : astropy.units.Quantity 
        The effective temperature of the spot umbra.
    Teff_penumbra : astropy.units.Quantity 
        The effective temperature of spot penumbra.
    r_A : float
        The ratio of total spot area to umbra area. 5+/-1 according
        to :cite:t:`2003A&ARv..11..153S`.
    growing : bool
        Whether or not the spot is growing.
    growth_rate : astropy.units.Quantity 
        Fractional growth of the spot for a given unit time.
        From from sunspot literature, can be 0.52/day to 1.83/day :cite:p:`2003A&ARv..11..153S`.
        According to M dwarf literature, can effectively be 0 :cite:p:`2015ApJ...806..212D`.
    decay_rate : astropy.units.Quantity
        The rate at which a spot linearly decays. From sunspot
        literature, this is 10.89 MSH/day :cite:p:`2003A&ARv..11..153S`. According to M dwarf
        literature, this can be 0 :cite:p:`2015ApJ...806..212D`.
    Nlat : int, default=500
        Number of latitude points.
    Nlon : int, default=1000
        Number of longitude points.
    gridmaker : CoordinateGrid, default=None
        Coordinate grid object to produce points in the
        stellar surface. Ideally, this is passed from a
        container object (such as `SpotCollection`).

    Attributes
    ----------
    radius
    coords : dict
        A dictionary containing the latitude and longitude of the spot's center.
    area_max : astropy.units.Quantity
        The maximum area a spot reaches before it decays.
    area_current : astropy.units.Quantity
        The current area of the spot.
    Teff_umbra : astropy.units.Quantity
        The effective temperature of the spot umbra.
    Teff_penumbra : astropy.units.Quantity
        The effective temperature of the spot penumbra.
    decay_rate : astropy.units.Quantity
        The rate at which a spot linearly decays.
    total_area_over_umbra_area : float
        The ratio of total spot area to umbra area. 5+/-1 according
        to :cite:t:`2003A&ARv..11..153S`.
    is_growing : bool
        Whether or not the spot is growing.
    growth_rate : astropy.units.Quantity
        Fractional growth of the spot for a given unit time.
    gridmaker : CoordinateGrid or None
        A `CoordinateGrid` object used to produce points on the stellar surface. If None,
        a `CoordinateGrid` object is created with default parameters.
    r : np.ndarray
        An array of points on the stellar surface with their pre-computed
        distance from the center of the spot.

    References
    ----------
    :cite:t:`2003A&ARv..11..153S`
    :cite:t:`2015ApJ...806..212D`
    """

    def __init__(
        self,
        lat: Quantity,
        lon: Quantity,
        area_max: Quantity,
        area_current: Quantity,
        teff_umbra: Quantity,
        teff_penumbra: Quantity,
        area_over_umbra_area: float = 5,
        is_growing: bool = True,
        growth_rate: Quantity = 0.52/u.day,
        decay_rate: Quantity = 10.89 * MSH/u.day,
        grid_params: Union[int,Tuple[int, int]] = (500, 1000),
        gridmaker: CoordinateGrid = None
    ):

        self.coords = {'lat': lat, 'lon': lon}
        self.area_max = area_max
        self.area_current = area_current
        self.teff_umbra = teff_umbra
        self.teff_penumbra = teff_penumbra
        self.decay_rate = decay_rate
        self.total_area_over_umbra_area = area_over_umbra_area
        self.is_growing = is_growing
        self.growth_rate = growth_rate

        if gridmaker is None:
            self.set_gridmaker(CoordinateGrid.new(grid_params))
        else:
            self.set_gridmaker(gridmaker)

    def set_gridmaker(self, gridmaker: CoordinateGrid):
        """
        Set the `gridmaker` attribute safely.

        Parameters
        ----------
        gridmaker : VSPEC.helpers.CoordinateGrid
            The `CoordinateGrid` object to set
        """
        self.gridmaker = gridmaker
        latgrid, longrid = self.gridmaker.grid()
        lat0 = self.coords['lat']
        lon0 = self.coords['lon']
        self.r = 2 * np.arcsin(np.sqrt(np.sin(0.5*(lat0-latgrid))**2
                                       + np.cos(latgrid)*np.cos(lat0)*np.sin(0.5*(lon0 - longrid))**2))

    def __str__(self):
        s = 'StarSpot with '
        s += f'Teff = ({self.teff_umbra:.0f},{self.teff_penumbra:.0f}), '
        s += f'area = {self.area_current.to(MSH):.0f}, '
        s += f'lat = {self.coords["lat"]:.1f}, lon = {self.coords["lon"]:.1f}'
        return s

    @property
    def radius(self) -> Quantity:
        """
        Radius

        Get the radius of the spot.

        Returns
        -------
        astropy.units.Quantity
            The radius of the spot.
        """
        return np.sqrt(self.area_current/np.pi).to(u.km)

    def angular_radius(self, star_rad: Quantity) -> Quantity:
        """
        Angular radius

        Get the angular radius of the spot on the stellar surface.

        Parameters
        ----------
        star_rad : astropy.units.Quantity 
            The radius of the star.

        Returns
        -------
        astropy.units.Quantity 
            The angular radius of the spot.
        """
        cos_angle = 1 - self.area_current/(2*np.pi*star_rad**2)
        return (np.arccos(cos_angle)).to(u.deg)

    def map_pixels(self, star_rad: Quantity) -> dict:
        """
        Map latitude and longituide points continaing the umbra and penumbra

        Parameters
        ----------
        star_rad : astropy.units.Quantity 
            The radius of the star.

        Returns
        -------
        dict
            Dictionary of points covered by the umbra and penumbra. Keys are the
            effective temperature (Teff) of each region. Values are numpy boolean
            arrays.
        Notes
        -----
        This function calculates the angular radius of the spot based on the
        provided `star_rad`, and uses the `total_area_over_umbra_area` attribute to
        determine the size of the spot. Then, it returns a dictionary with two
        keys, `Teff_umbra` and `Teff_penumbra`, whose values are boolean arrays
        indicating which points are covered by each region.
        """
        radius = self.angular_radius(star_rad)
        radius_umbra = radius/np.sqrt(self.total_area_over_umbra_area)
        return {self.teff_umbra: self.r < radius_umbra,
                self.teff_penumbra: self.r < radius}

    def surface_fraction(self, sub_obs_coords: dict,
                         star_rad: Quantity, n_points: int = 1001) -> float:
        """
        Determine the surface fraction covered by a spot from a given 
        angle of observation using the orthographic projection.

        Parameters
        ----------
        sub_obs_coords : dict
            Dictionary giving coordinates of the sub-observation point. This
            is the point that is at the center of the stellar disk from the
            view of an observer. Format: {'lat': lat, 'lon': lon} where lat and
            lon are astropy.units.Quantity objects.
        star_rad : astropy.units.Quantity 
            Radius of the star.
        N : int, optional
            Number of points to use in numerical integration.
            N=1000 is not so different from N=100000.

        Returns
        -------
        float
            Fraction of observed disk covered by the spot.

        """
        cos_c0 = (np.sin(sub_obs_coords['lat']) * np.sin(self.coords['lat'])
                  + np.cos(sub_obs_coords['lat']) * np.cos(self.coords['lat'])
                  * np.cos(sub_obs_coords['lon']-self.coords['lon']))
        c0 = np.arccos(cos_c0)
        c = np.linspace(-90, 90, n_points)*u.deg
        a = self.angular_radius(star_rad).to(u.deg)
        rad = a**2 - (c-c0)**2
        rad[rad < 0] = 0
        integrand = 2 * np.cos(c)*np.sqrt(rad)
        return (np.trapz(integrand, x=c)/(2*np.pi*u.steradian)).to_value(u.dimensionless_unscaled)

    def age(self, time: Quantity) -> None:
        """
        Age a spot according to its growth timescale and decay rate

        Parameters
        ----------
        time : astropy.units.Quantity
            Length of time to age the spot. For most realistic behavior,
            time should be << spot lifetime

        Notes
        -----
        This method updates the `area_current` attribute of the `StarSpot`
        object to simulate the growth and decay of a sunspot.

        .. math::
            A(t) = \\left\\{
                \\begin{array}{lr}
                    A_0 e^{(t-t_0)/\\tau}, & \\text{if } t \\leq t_0 \\\\
                    A_0 - W(t-t_0), & \\text{if } t > t_0
                \\end{array}
                \\right\\}


        If the spot is growing, it calculates the time required to reach
        the maximum area (`time_to_max`) using the growth rate (`growth_rate`)
        and the maximum area (`area_max`). If `time_to_max` is greater than
        the input time (`time`), it grows the spot and updates the `area_current` 
        attribute. If `time_to_max` is less than or equal to the input time, it
        sets the `is_growing` attribute to False and calculates the area that the
        spot should decay (`area_decay`). If the decayed area (`area_decay`) is
        greater than the maximum area (`area_max`), the `area_current` attribute
        is set to zero. Otherwise, it updates the `area_current` attribute accordingly.

        If the spot is not growing, it calculates the area that the spot should
        decay (`area_decay`). If the decayed area (`area_decay`) is greater than
        the maximum area (`area_max`), the `area_current` attribute is set to zero.
        Otherwise, it updates the `area_current` attribute accordingly.
        """
        if self.is_growing:
            tau = np.log((self.growth_rate * u.day).to(u.Unit('')) + 1)
            if tau == 0:
                time_to_max = np.inf*u.day
            else:
                time_to_max = np.log(
                    self.area_max/self.area_current)/tau * u.day
            if time_to_max > time:
                new_area = self.area_current * np.exp(tau * time/u.day)
                self.area_current = new_area
            else:
                self.is_growing = False
                decay_time = time - time_to_max
                area_decay = decay_time * self.decay_rate
                if area_decay > self.area_max:
                    self.area_current = 0*MSH
                else:
                    self.area_current = self.area_max - area_decay
        else:
            area_decay = time * self.decay_rate
            if area_decay > self.area_max:
                self.area_current = 0*MSH
            else:
                self.area_current = self.area_current - area_decay


class SpotCollection:
    """
    Container holding StarSpot objects

    Parameters
    ----------
    *spots : tuple of StarSpot objects
        A series of StarSpot objects to be added to the collection.
    Nlat : int, default=500
        Number of latitude points.
    Nlon : int, default=1000
        Number of longitude points.
    gridmaker : CoordinateGrid, default=None
        `CoordinateGrid` object, probably passed from a
        `Star`

    Notes
    -----
    This class is a container for `StarSpot` objects. It can be
    used to store a series of spots and to apply operations to the entire collection.

    Attributes
    ----------
    spots : tuple of StarSpot objects
        Series of `StarSpot` objects in the collection.
    gridmaker : CoordinateGrid object
        `CoordinateGrid` object used to calculate the grid of the stellar surface.
    """
    gridmaker: CoordinateGrid = None
    def __init__(
        self,
        *spots: StarSpot,
        grid_params: Union[int,Tuple[int, int]] = (500, 1000),
        gridmaker: CoordinateGrid=None
    ):
        self.spots = spots
        if gridmaker is None:
            gridmaker = CoordinateGrid.new(grid_params)
        self.set_gridmaker(gridmaker)
    def set_gridmaker(self, gridmaker: CoordinateGrid):
        """
        Set the `gridmaker` attribute safely.

        Parameters
        ----------
        gridmaker : VSPEC.helpers.CoordinateGrid
            The `CoordinateGrid` object to set
        """
        self.gridmaker = gridmaker
        for spot in self.spots:
            spot.set_gridmaker(self.gridmaker)

    def add_spot(self, spot: Typing.Union[StarSpot, list[StarSpot]]):
        """
        Add a `StarSpot` object or a list of `StarSpot` objects to the collection.

        Parameters
        ----------
        spot : `StarSpot` or list of `StarSpot` objects
            `StarSpot` object(s) to be added to the collection.

        Notes
        -----
        This method adds a `StarSpot` object or a list of `StarSpot`
        objects to the `spots` attribute of the `SpotCollection` object.

        For every spot added to the `spots` attribute, each first has it's
        own `gridmaker` attribute set to be identical to the `SpotColelction`'s
        own `gridmaker` attribute.
        """
        if isinstance(spot, StarSpot):
            spot = [spot]
        for s in spot:
            s.gridmaker = self.gridmaker
        self.spots += tuple(spot)

    def clean_spotlist(self):
        """
        Remove spots that have decayed to 0 area from
        the `spots` attribute of the `SpotCollection` object.

        Notes
        -----
        This method iterates over the `StarSpot` objects in the `spots`
        attribute of the `SpotCollection` object and removes those that
        have decayed to 0 area and are not growing. The remaining
        `StarSpot` objects are stored back in the `spots` attribute of
        the `SpotCollection` object.
        """
        spots_to_keep = []
        for spot in self.spots:
            if (spot.area_current <= 0*MSH) and (not spot.is_growing):
                pass
            else:
                spots_to_keep.append(spot)
        self.spots = spots_to_keep

    def map_pixels(self, star_rad: Quantity, star_teff: Quantity):
        """
        Map latitude and longitude points containing the umbra and penumbra
        of each spot. For pixels with coverage from multiple spots, assign
        the coolest Teff to that pixel.

        Parameters
        ----------
        latgrid : astropy.units.Quantity , shape(M,N)
            Grid of latitude points to map.
        longrid : astropy.units.Quantity , shape(M,N)
            Grid of longitude points to map.
        star_rad : astropy.units.Quantity
            Radius of the star.
        star_teff : astropy.units.Quantity
            Temperature of the star.

        Returns
        -------
        surface_map : array of astropy.units.Quantity , shape(M,N)
            Map of the stellar surface with Teff assigned to each pixel
        """
        surface_map = self.gridmaker.zeros()*star_teff.unit + star_teff
        for spot in self.spots:
            teff_dict = spot.map_pixels(star_rad)
            # penumbra
            penumbra = teff_dict[spot.teff_penumbra] & (
                surface_map > spot.teff_penumbra)
            # umbra
            umbra = teff_dict[spot.teff_umbra] & (
                surface_map > spot.teff_umbra)
            surface_map = np.where(umbra, spot.teff_umbra, np.where(
                penumbra, spot.teff_penumbra, surface_map))
        return surface_map

    def age(self, time: Quantity) -> None:
        """
        Age spots according to its growth timescale and decay rate.

        Remove spots that have decayed.

        Parameters
        ----------
        time : astropy.units.Quantity
            Length of time to age the spot. For most realistic
            behavior, time should be << spot lifetime.
        """
        for spot in self.spots:
            spot.age(time)
        self.clean_spotlist()

    def get_coverage(
        self,
        r_star: u.Quantity,
    ):
        """
        Get the fractional coverage of star spots.

        Parameters
        ----------
        r_star : astropy.units.Quantity
            The radius of the star.

        Returns
        -------
        coverage : float
            The fraction of the stellar surface covered by spots.
        """
        teff_star = np.inf*u.K
        area = self.gridmaker.area
        tmap = self.map_pixels(r_star, teff_star)
        is_spot = tmap != teff_star
        return np.sum(area*is_spot)/np.sum(area)


class SpotGenerator:
    """Spot Generator

    Class controling the birth rates and properties of new spots.
    This class is based on various studies, but since in general starspots cannot
    be resolved, lots of gaps are filled in with studies of sunspots.

    Parameters
    ----------
    average_area : astropy.units.Quantity 
        The average peak spot area.
    area_spread : float
        The standard deviation, in dex, of the lognormal peak spot area distribution
    umbra_teff : astropy.units.Quantity 
        Effective temperature of umbra.
    penumbra_teff : astropy.units.Quantity 
        Effective temperature of penumbra.
    growth_rate : astropy.units.Quantity , default=0.52/u.day
        The spot growth rate.
    decay_rate : astropy.units.Quantity [area/time], default = 10.89 * MSH/u.day
        The spot decay rate.
    starting_size : astropy.units.Quantity , default=10*MSH
        The area of each spot at birth.
    distribution : str, default='solar'
        The spot distribution method. 'iso' or 'solar'.
    coverage : float, default=0.2
        The factional coverage of surface by spots in growth-decay equillibrium.
    Nlat : int, default=500
        The number of latitude points on the stellar sufrace.
    Nlon : int, default=1000
        The number of longitude points on the stellar surface.
    gridmaker : CoordinateGrid, default=None
        A `CoordinateGrid` object to create the stellar sufrace grid.
    rng : numpy.random.generator, default = numpy.random.default_rng()
        The random number generator to use.

    Attributes
    ----------
    average_spot_area : astropy.units.Quantity
        The average peak spot area.
    spot_area_spread : float
        The standard deviation, in dex, of the lognormal peak spot area distribution.
    umbra_teff : astropy.units.Quantity
        Effective temperature of umbra.
    penumbra_teff : astropy.units.Quantity
        Effective temperature of penumbra.
    growth_rate : astropy.units.Quantity
        The spot growth rate.
    decay_rate : astropy.units.Quantity
        The spot decay rate.
    starting_size : astropy.units.Quantity
        The area of each spot at birth.
    distribution : str
        The spot distribution method. Choose from 'iso' or 'solar'.
    coverage : float
        The fractional coverage of the surface by spots in growth-decay equilibrium.
    gridmaker : CoordinateGrid or subclass
        A `CoordinateGrid` object to create the stellar surface grid.
    rng : numpy.random.generator
        The random number generator to use.
    is_static
    mean_lifetime
    mean_area

    Notes
    -----
    The `distribution` parameter can have values of 'iso' of 'solar'. The 'iso' keyword
    distributes spots evenly across the surface. The 'solar' keyword, on the other hand,
    distributes spots according to their know clustering around +/- 15 degrees solar
    latitude [1]_.

    References
    ----------
    .. [1] Mandal, S., Karak, B. B., & Banerjee, D. 2017, ApJ, 851, 70

    """

    def __init__(self,
                 dist_area_mean: Quantity,
                 dist_area_logsigma: float,
                 umbra_teff: Quantity,
                 penumbra_teff: Quantity,
                 growth_rate: Quantity = 0.52/u.day,
                 decay_rate: Quantity = 10.89 * MSH/u.day,
                 init_area: Quantity = 10*MSH,
                 distribution='solar',
                 coverage: float = 0.2,
                 grid_params: Union[int,Tuple[int, int]] = (500, 1000),
                 gridmaker:CoordinateGrid=None,
                 rng: np.random.Generator = np.random.default_rng()
                 ):
        self.dist_area_mean = dist_area_mean
        self.dist_area_logsigma = dist_area_logsigma
        self.umbra_teff = umbra_teff
        self.penumbra_teff = penumbra_teff
        self.growth_rate = growth_rate
        self.decay_rate = decay_rate
        self.init_area = init_area
        self.distribution = distribution
        self.coverage = coverage
        self.grid_params = grid_params
        if gridmaker is None:
            self.gridmaker = CoordinateGrid.new(grid_params)
        else:
            self.gridmaker = gridmaker
        self.rng = rng
    @classmethod
    def off(
        cls,
        grid_params: Union[int, Tuple[int, int]] = 1000,
        gridmaker: CoordinateGrid = None,
        rng: np.random.Generator = np.random.default_rng()
    ):
        """
        Create an instance that does not generate anything.
        
        Parameters
        ----------
        grid_params : int or tuple
            The number of latitude and longitude points on the stellar surface.
        gridmaker : CoordinateGrid
            A `CoordinateGrid` object to create the stellar surface grid.
        rng : numpy.random.generator
            The random number generator to use.
        """
        return cls(
            dist_area_mean=10*MSH,
            dist_area_logsigma=1,
            umbra_teff=0*u.K,
            penumbra_teff=0*u.K,
            growth_rate=0*u.day,
            decay_rate=0*MSH/u.day,
            init_area=10*MSH,
            distribution='iso',
            coverage=0,
            grid_params=grid_params,
            gridmaker=gridmaker,
            rng=rng
        )
        

    @property
    def is_static(self) -> bool:
        """
        True if the spots do not decay.

        Returns
        -------
        is_static : bool
            True if the spots do not decay.
        """
        return self.decay_rate == 0*MSH/u.day

    @property
    def mean_lifetime(self) -> u.Quantity:
        """
        The mean lifetime of the spots.

        Returns
        -------
        mean_lifetime : astropy.units.Quantity
            The mean lifetime of spots.
        """
        if self.is_static:
            return np.inf * u.day
        else:
            return ((self.dist_area_mean/self.decay_rate)
                    - np.log(self.init_area/self.dist_area_mean)/self.growth_rate).to(u.day)

    @property
    def mean_area(self) -> u.Quantity:
        """
        The time-averaged area of a typical spot.

        Returns
        -------
        mean_area : astropy.units.Quantity
            The time-averaged area of a typical spot.
        """
        if self.is_static:
            return self.dist_area_mean
        else:
            integrated_area = (self.dist_area_mean / self.growth_rate) \
                + (self.init_area / self.growth_rate) \
                + (self.dist_area_mean**2/2/self.decay_rate)
            return integrated_area / self.mean_lifetime

    def get_coordinates(self, n_spots: int):
        """
        Get coordinates for a `StarSpot` distribution.

        Parameters
        ----------
        n_spots : int
            Number of spots to create.

        Returns
        -------
        lat : astropy.units.Quantity
            Latitude coordinates of the new spots
        lon : astropy.units.Quantity
            Longitude coordinates of the new spots

        Notes
        -----
        In order to draw latitude points for the isotropic case, we
        use inverse transform sampling to account for the jacobian.
        """
        if self.distribution == 'solar':
            # (dist approx from 2017ApJ...851...70M)
            hemi = self.rng.choice([-1, 1], size=n_spots)
            lat = self.rng.normal(15, 5, size=n_spots)*hemi*u.deg
            lon = self.rng.random(size=n_spots)*360*u.deg
        elif self.distribution == 'iso':
            lon = self.rng.random(size=n_spots)*360*u.deg
            # use inverse transform to generate lats
            x = self.rng.random(size=n_spots)
            lat = np.arcsin(2*x - 1)/np.pi * 180*u.deg
        else:
            raise ValueError(
                f'Unknown value {self.distribution} for distribution')
        return lat, lon

    def generate_spots(self, n_spots: int) -> tuple[StarSpot]:
        """
        Create a specified number of `StarSpot` objects.

        Parameters
        ----------
        n_spots: int
            Number of spots to create.

        Returns
        -------
        tuple[StarSpot]
            Tuple of new `StarSpot` objects.

        Raises
        ------
        ValueError
            If an unknown value is given for distribution.
        """
        new_max_areas = self.rng.lognormal(mean=np.log(
            self.dist_area_mean/MSH), sigma=self.dist_area_logsigma, size=n_spots)*MSH
        new_area_ratio = self.rng.normal(loc=5, scale=1, size=n_spots)
        while np.any(new_area_ratio <= 0):
            new_area_ratio = self.rng.normal(loc=5, scale=1, size=n_spots)
        lat, lon = self.get_coordinates(n_spots)

        penumbra_teff = self.penumbra_teff
        umbra_teff = self.umbra_teff

        spots = []
        for i in range(n_spots):
            spots.append(StarSpot(
                lat[i], lon[i],
                new_max_areas[i],
                self.init_area,
                umbra_teff, penumbra_teff,
                growth_rate=self.growth_rate,
                decay_rate=self.decay_rate,
                area_over_umbra_area=new_area_ratio[i],
                grid_params = self.grid_params,
                gridmaker=self.gridmaker
            ))
        return tuple(spots)

    def get_n_spots_to_birth(self, time: Quantity, rad_star: Quantity) -> float:
        """
        Calculate how many new `StarSpot` objects to birth over
        a given time duration (expectation value).

        Parameters
        ----------
        time : astropy.units.Quantity 
            Amount of time in which to birth spots.
            The total number of new spots will consider this time and the birthrate.
        rad_star : astropy.units.Quantity 
            The radius of the star.

        Returns
        -------
        N_exp : float
            Expected number of new `StarSpot` objects.
        """
        n_expected = (self.coverage * 4*np.pi*rad_star**2 / self.mean_area *
                 time / self.mean_lifetime).to_value(u.dimensionless_unscaled)
        return n_expected

    def birth_spots(self, time: Quantity, rad_star: Quantity) -> tuple[StarSpot]:
        """
        Generate new `StarSpot` objects to be birthed over a given time duration.

        Parameters
        ----------
        time : astropy.units.Quantity 
            Amount of time in which to birth spots.
            The total number of new spots will consider this time and the birthrate.
        rad_star : astropy.units.Quantity 
            The radius of the star.

        Returns
        -------
        Tuple[StarSpot]
            New `StarSpot` objects.
        """
        n_expected = self.get_n_spots_to_birth(time, rad_star)
        # N_exp is the expectation value of N, but this is a poisson process
        # N = max(0, round(np.random.normal(loc=N_exp, scale=np.sqrt(N_exp))))
        n_actual = self.rng.poisson(lam=n_expected)

        return self.generate_spots(n_actual)

    def generate_mature_spots(self, coverage: float, r_star: Quantity) -> List[StarSpot]:
        """Generate mature StarSpot objects to cover a given fraction of the star's surface.

        This method generates mature spots such that the total solid angle subtended by the spots
        covers a specified fraction of the star's surface.

        Parameters
        ----------
        coverage : float
            The fraction of the star's surface to be covered by the spots.
        r_star : astropy.units.Quantity 
            The radius of the star.

        Returns:
            List[StarSpot]: A list of mature spots generated by this method.

        Raises:
            ValueError: If the coverage is greater than 1 or less than 0.
        """
        if coverage > 1 or coverage < 0:
            raise ValueError('Coverage must be between 0 and 1.')
        spots = []
        current_omega = 0*(u.deg**2)
        target_omega = (4*np.pi*coverage*u.steradian).to(u.deg**2)
        while current_omega < target_omega:
            new_spot = self.generate_spots(1)[0]
            if self.is_static:
                area0 = self.init_area
                area_range = new_spot.area_max - area0
                area = self.rng.random()*area_range + area0
                new_spot.area_current = area
            else:
                age = self.rng.random() * self.mean_lifetime
                new_spot.age(age)
            spots.append(new_spot)
            spot_solid_angle = new_spot.angular_radius(r_star)**2 * np.pi
            current_omega += spot_solid_angle
        return spots



"""
Tests for vspec_vsm.spots.StarSpot
"""
from astropy import units as u
import numpy as np
import pytest

from vspec_vsm.spots import StarSpot, SpotCollection, SpotGenerator
from vspec_vsm.coordinate_grid import RectangularGrid
from vspec_vsm.config import MSH

rng = np.random.default_rng(20)


def init_test_spot(**kwargs):
    """
    Initialize a default StarSpot for testing
    """
    lat = kwargs.get('lat', 0*u.deg)
    lon = kwargs.get('lon', 0*u.deg)
    Amax = kwargs.get('Amax', 500*MSH)
    A0 = kwargs.get('A0', 400*MSH)
    Teff_umbra = kwargs.get('Teff_umbra', 2700*u.K)
    Teff_penumbra = kwargs.get('Teff_penumbra', 2500*u.K)
    r_A = kwargs.get('r_A', 5.)
    growing = kwargs.get('growing', True)
    growth_rate = kwargs.get('growth_rate', 0/u.day)
    decay_rate = kwargs.get('decay_rate', 0*MSH/u.day)
    grid_params = kwargs.get('grid_params', 1000)
    gridmaker = kwargs.get('gridmaker', None)

    spot = StarSpot(
        lat=lat,
        lon=lon,
        area_max=Amax,
        area_current=A0,
        teff_umbra=Teff_umbra,
        teff_penumbra=Teff_penumbra,
        area_over_umbra_area=r_A,
        is_growing=growing,
        growth_rate=growth_rate,
        decay_rate=decay_rate,
        grid_params=grid_params,
        gridmaker=gridmaker
    )
    return spot


def test_spot_init():
    """
    Test __init__() for starspots
    """
    spot = init_test_spot(grid_params=20000)
    assert np.all(spot.r > 0*u.deg)
    assert np.all(spot.r < 180*u.deg)
    assert np.max(spot.r.to_value(u.deg)) == pytest.approx(180, abs=1)

    other = init_test_spot(
        gridmaker=RectangularGrid(500+1, 1000),
    )
    assert spot.gridmaker != other.gridmaker


def test_spot_str():
    """
    Make sure the __str__ method works.
    """
    spot = init_test_spot()
    s = str(spot)
    assert isinstance(s, str)


def test_spot_radius():
    """
    Test the spot radius
    """
    current_area = 400*MSH
    spot = init_test_spot(A0=current_area)
    # A = pi r**2
    # r = sqrt(A/pi)
    pred = np.sqrt(current_area/np.pi)
    obs = spot.radius
    assert pred.to_value(u.km) == pytest.approx(obs.to_value(u.km), rel=1e-6)


def test_spot_angular_radius():
    """
    Tests for the angular radius of the spot.
    """
    stellar_rad = 1000*u.km
    star_surface_area = 4*np.pi*stellar_rad**2
    spot = init_test_spot(A0=star_surface_area)

    pred = 180*u.deg
    obs = spot.angular_radius(stellar_rad)
    assert obs.to_value(u.deg) == pytest.approx(
        pred.to_value(u.deg), abs=0.01)

    spot = init_test_spot(A0=0.5*star_surface_area)
    pred = 90*u.deg
    obs = spot.angular_radius(stellar_rad)
    assert obs.to_value(u.deg) == pytest.approx(
        pred.to_value(u.deg), abs=0.01)


def test_spot_map_pixels():
    """
    Test StarSpot.map_pixels
    """
    stellar_rad = 1000*u.km
    star_surface_area = 4*np.pi*stellar_rad**2
    spot = init_test_spot(A0=star_surface_area, r_A=1)  # no penumbra
    pixmaps = spot.map_pixels(stellar_rad)
    umbra = pixmaps[spot.teff_umbra]
    penumbra = pixmaps[spot.teff_penumbra]
    assert np.all(umbra)
    assert not np.any(penumbra & ~umbra)

    spot = init_test_spot(A0=star_surface_area, r_A=np.inf)  # no umbra
    pixmaps = spot.map_pixels(stellar_rad)
    umbra = pixmaps[spot.teff_umbra]
    penumbra = pixmaps[spot.teff_penumbra]
    assert np.all(penumbra)
    assert not np.any(umbra)

    spot = init_test_spot(A0=star_surface_area, r_A=4, lat=0*u.deg)
    # half and half --> This is a sphere, so our 2D approximation goes out the window
    # for a large radius spot.
    pixmaps = spot.map_pixels(stellar_rad)
    umbra = pixmaps[spot.teff_umbra]
    penumbra = pixmaps[spot.teff_penumbra]
    lats, _ = spot.gridmaker.grid()
    sin_theta = np.cos(lats)
    assert (umbra*sin_theta.value).sum() == pytest.approx(((penumbra &
                                                            ~umbra)*sin_theta.value).sum(), rel=0.01)

    spot = init_test_spot(A0=1e-3*star_surface_area,
                          r_A=2, grid_params=(2000,4000))
    # half and half --> we try to get close to 2D without the discrete grid messing things up
    pixmaps = spot.map_pixels(stellar_rad)
    umbra = pixmaps[spot.teff_umbra]
    penumbra = pixmaps[spot.teff_penumbra]
    lats, _ = spot.gridmaker.grid()
    sin_theta = np.cos(lats)
    assert (umbra*sin_theta.value).sum() == pytest.approx(
        ((penumbra & ~umbra)*sin_theta.value).sum(), rel=0.05)


def test_spot_surface_fraction():
    """
    Test StarSpot.sufrace_fraction
    """
    stellar_rad = 1000*u.km
    star_surface_area = 4*np.pi*stellar_rad**2
    spot = init_test_spot(A0=0.01*star_surface_area)
    sub_obs = {'lat': 0*u.deg, 'lon': 0*u.deg}
    assert spot.surface_fraction(
        sub_obs, stellar_rad) == pytest.approx(0.02, rel=0.01)
    sub_obs = {'lat': 0*u.deg, 'lon': 180*u.deg}
    assert spot.surface_fraction(
        sub_obs, stellar_rad) == pytest.approx(0.0, rel=0.01)


def test_spot_age():
    """
    Test StarSpot.age
    """
    spot = init_test_spot(A0=10*MSH, Amax=100*MSH, growth_rate=1 /
                          u.day, decay_rate=10*MSH/u.day, growing=True)
    step = 1*u.day
    assert spot.area_current.to_value(MSH) == pytest.approx(10, rel=0.01)
    spot.age(step)
    assert spot.area_current.to_value(MSH) == pytest.approx(20, rel=0.01)
    spot.age(step)
    assert spot.area_current.to_value(MSH) == pytest.approx(40, rel=0.01)
    spot.age(step)
    assert spot.area_current.to_value(MSH) == pytest.approx(80, rel=0.01)
    spot.age(step)
    assert not spot.is_growing

    spot = init_test_spot(A0=100*MSH, Amax=100*MSH, growth_rate=1 /
                          u.day, decay_rate=10*MSH/u.day, growing=False)
    assert spot.area_current.to_value(MSH) == pytest.approx(100, rel=0.01)
    spot.age(step)
    assert spot.area_current.to_value(MSH) == pytest.approx(90, rel=0.01)
    spot.age(9*step)
    assert spot.area_current.to_value(MSH) == pytest.approx(0, rel=0.01)

    spot = init_test_spot(A0=10*MSH, Amax=100*MSH, growth_rate=1 /
                          u.day, decay_rate=10*MSH/u.day, growing=True)
    step = 1*u.day
    assert spot.area_current.to_value(MSH) == pytest.approx(10, rel=0.01)
    spot.age(20*step)
    assert spot.area_current.to_value(MSH) == pytest.approx(0, rel=0.01)


def test_init_spot_collection():
    """
    Test spot initialization
    """
    N = 4
    collec = SpotCollection(*[init_test_spot()
                            for i in range(N)], grid_params=(300, 600))
    expected_grid = RectangularGrid(300, 600)
    for spot in collec.spots:
        assert isinstance(spot, StarSpot)
    assert collec.gridmaker == expected_grid
    for spot in collec.spots:
        assert spot.gridmaker == collec.gridmaker


def test_spot_collection_add_spot():
    """
    Test `SpotCollection.add_spot()`
    """
    N = 4
    collec = SpotCollection(*[init_test_spot()
                            for i in range(N)], grid_params=(300, 600))
    assert len(collec.spots) == N
    new_spot = init_test_spot()
    collec.add_spot(new_spot)
    assert len(collec.spots) == N+1
    new_spots = [init_test_spot() for i in range(N)]
    collec.add_spot(new_spots)
    assert len(collec.spots) == 2*N+1
    for spot in collec.spots:
        assert spot.gridmaker == collec.gridmaker


def test_spot_collection_clean_spotlist():
    """
    Test `SpotCollection.clean_spotlist()`
    """
    N = 4
    spots = [init_test_spot() for i in range(N)] + \
        [init_test_spot(A0=0*MSH, growing=True)]
    collec = SpotCollection(*spots, grid_params=(300, 600))
    collec.clean_spotlist()
    assert len(collec.spots) == N+1
    spots = [init_test_spot() for i in range(N)] + \
        [init_test_spot(A0=0*MSH, growing=False)]
    collec = SpotCollection(*spots, grid_params=(300, 600))
    collec.clean_spotlist()
    assert len(collec.spots) == N


def test_spot_collection_map_pixels():
    """
    Test `SpotCollection.map_pixels()`
    """
    R_star = 0.15*u.R_sun
    Teff = 3300*u.K
    spots = [
        init_test_spot(Teff_umbra=2500*u.K, Teff_penumbra=2500*u.K),
        init_test_spot(Teff_umbra=2700*u.K, Teff_penumbra=2700*u.K)
    ]
    collec = SpotCollection(*spots, grid_params=(300, 600))
    pmap = collec.map_pixels(R_star, Teff)
    assert not np.any(pmap == 2700*u.K)

    spots = [
        init_test_spot(Teff_umbra=3500*u.K, Teff_penumbra=3500*u.K),
        init_test_spot(Teff_umbra=3700*u.K, Teff_penumbra=3700*u.K)
    ]
    collec = SpotCollection(*spots, grid_params=(300, 600))
    pmap = collec.map_pixels(R_star, Teff)
    assert np.all(pmap == Teff)


def test_spot_collection_age():
    """
    Test `SpotCollection.age()`
    """
    spot = init_test_spot(A0=10*MSH, Amax=100*MSH, growth_rate=1 /
                          u.day, decay_rate=10*MSH/u.day, growing=True)
    time = 1*u.day
    collec = SpotCollection(spot, grid_params=(300, 600))
    assert collec.spots[0].area_current == 10*MSH
    collec.age(time)
    assert collec.spots[0].area_current == 20*MSH
    collec.age(time*20)
    assert len(collec.spots) == 0


def init_spot_generator(**kwargs):
    """
    Initialize a `SpotGenerator` for testing
    """
    return SpotGenerator(
        dist_area_mean=kwargs.get('average_area', 100*MSH),
        dist_area_logsigma=kwargs.get('area_spread', 0.1),
        umbra_teff=kwargs.get('umbra_teff', 2500*u.K),
        penumbra_teff=kwargs.get('penumbra_teff', 2700*u.K),
        growth_rate=kwargs.get('growth_rate', 0/u.day),
        decay_rate=kwargs.get('decay_rate', 0*MSH/u.day),
        init_area=kwargs.get('starting_size', 10*MSH),
        distribution=kwargs.get('distribution', 'iso'),
        coverage=kwargs.get('coverage', 0.0),
        grid_params=kwargs.get('grid_params', (300, 600)),
        rng=rng
    )


def test_spot_generator_init():
    """
    Test `SpotGenerator.__init__()`
    """
    Nlat, Nlon = 300, 600
    gen = init_spot_generator(grid_params=(Nlat, Nlon))
    grid = RectangularGrid(Nlat, Nlon)
    assert gen.gridmaker == grid


def test_spot_generator_is_static():
    gen = init_spot_generator(growth_rate=0/u.day, decay_rate=0*MSH/u.day)
    assert gen.is_static
    gen = init_spot_generator(growth_rate=0/u.day, decay_rate=10*MSH/u.day)
    assert not gen.is_static


def test_spot_generator_mean_lifetime():
    growth_rate = 0.5/u.day
    decay_rate = 10*MSH/u.day
    max_size = 100*MSH
    starting_size = 10*MSH
    gen = init_spot_generator(growth_rate=growth_rate, decay_rate=decay_rate,
                              average_area=max_size, starting_size=starting_size)
    expected_value = max_size/decay_rate - \
        np.log(starting_size/max_size)/growth_rate
    assert expected_value == gen.mean_lifetime


def test_spot_generator_N_spots_to_birth():
    """
    Test `SpotGenerator.get_N_spots_to_birth()`
    """
    time = 10*u.day
    r_star = 1*u.R_sun
    gen = init_spot_generator(average_area=100*MSH,
                              coverage=0.0, decay_rate=10*MSH/u.day, growth_rate=0.5/u.day)
    exp = 0.0
    assert gen.get_n_spots_to_birth(
        time, r_star) == pytest.approx(exp, abs=1e-6)


def test_spot_generator_get_coordinates():
    """
    Test `SpotGenerator.get_coordinates()`
    """
    N = 3
    gen = init_spot_generator(distribution='iso')
    lats, lons = gen.get_coordinates(N)
    assert len(lats) == N
    assert lats.unit == u.deg
    assert len(lons) == N
    assert lons.unit == u.deg

    gen = init_spot_generator(distribution='even')
    with pytest.raises(ValueError):
        gen.get_coordinates(N)


def test_spot_generator_generate_spots():
    """
    Test `SpotGenerator.generate_spots()`
    """
    N = 3
    gen = init_spot_generator(starting_size=10*MSH)
    spots = gen.generate_spots(N)
    assert len(spots) == N
    for spot in spots:
        assert spot.area_current == 10*MSH


def test_spot_generator_generate_mature_spots():
    """
    Test `SpotGenerator.generate_mature_spots()`
    """
    r_star = 0.15*u.R_sun
    starting_size = 10*MSH
    mean_spot_area = 500*MSH
    gen = SpotGenerator(
        dist_area_mean=mean_spot_area,
        dist_area_logsigma=0.2,
        umbra_teff=2600*u.K,
        penumbra_teff=2900*u.K,
        growth_rate=0/u.day,
        decay_rate=0*MSH/u.day,
        init_area=starting_size,
        distribution='iso',
        coverage=0.2,
        grid_params=(500, 1000), rng=rng
    )
    with pytest.raises(ValueError):
        gen.generate_mature_spots(-0.1, r_star)
    with pytest.raises(ValueError):
        gen.generate_mature_spots(1.1, r_star)
    spots = gen.generate_mature_spots(0.1, r_star)
    assert np.any([spot.area_current > starting_size for spot in spots])

    gen = init_spot_generator(
        starting_size=starting_size, growth_rate=1/u.day, decay_rate=10*MSH/u.day)
    spots = gen.generate_mature_spots(0.1, r_star)
    assert np.any([spot.area_current > starting_size for spot in spots])



"""
Most of the interface is defined in by the ``Star`` class.
"""

__verion__ = "0.3.0"

from vspec_vsm.star import Star
from vspec_vsm.spots import SpotCollection, StarSpot, SpotGenerator
from vspec_vsm.faculae import FaculaCollection, Facula, FaculaGenerator
from vspec_vsm.flares import StellarFlare, FlareCollection, FlareGenerator
from vspec_vsm.granules import Granulation, GranulationKernel
from vspec_vsm.config import MSH
from vspec_vsm.coordinate_grid import CoordinateGrid



"""
Configurations
"""

from astropy import units as u
import numpy as np

MSH = u.def_unit('msh', 1e-6 * 0.5 * 4*np.pi*u.R_sun**2)
"""
Micro-solar hemisphere

This is a standard unit in heliophysics that
equals one millionth of one half the surface area of the Sun.

:type: astropy.units.Unit
"""

stellar_area_unit = MSH
"""
The standard stellar surface area unit.

This unit is used to represent the surface area of stars in VSPEC.
The micro-solar hemisphere is chosen because most Sun Spot literature uses
this unit.

:type: astropy.units.Unit
"""

starspot_initial_area = 10*MSH
"""
Initial ``StarSpot`` area.

Because spots grow exponentially, they can't start at 0 area.
When they are born they are given this small area.

:type: astropy.units.Quantity

.. todo::
    This should optionaly be set by the user. So that smaller
    star spot area regimes are accessible.
"""

NLAT = 500
"""
The default latitude resolution for the stellar model. This should
be set by finding a balance between noticing small changes in spots/faculae
and computation time.

:type: int
"""

NLON = 1000
"""
The default longitude resolution for the stellar model. This should
be set by finding a balance between noticing small changes in spots/faculae
and computation time.

:type: int
"""

grid_teff_bounds = (2300*u.K, 3900*u.K)
"""
The limits on the effective temperature allowed by the grid.

:type: tuple of astropy.units.Quantity
"""



"""

Granulation
-----------

"""
import jax
import jax.numpy as jnp
from tinygp import kernels, GaussianProcess
from astropy import units as u
import numpy as np

time_unit = u.hr


class GranulationKernel(kernels.Kernel):
    """
    GP Kernel to describe Granulation

    Parameters
    ----------
    scale : float
        A scalar coefficient for the kernel function.
    period : float
        The period of the kernel in hours.

    Notes
    -----
    This kernel is based on [1]_.

    References
    ----------
    .. [1] Gordon, T. A., Agol, E., & Foreman-Mackey, D. 2020, AJ, 160, 240
    """

    def __init__(self, scale, period):
        self.scale = jnp.atleast_1d(scale)
        self.freq = 1/jnp.atleast_1d(period)

    def evaluate(self, X1, X2):
        tau = jnp.atleast_1d(jnp.abs(X1 - X2))[..., None]

        return jnp.sum(self.scale*self.freq*jnp.exp(
            -self.freq*tau/jnp.sqrt(2)
        )
            * jnp.cos((self.freq*tau/jnp.sqrt(2))-(jnp.pi/4)))


def build_gp(params: dict, X: np.ndarray) -> GaussianProcess:
    """
    Build Gaussian Process

    Parameters
    ----------
    params : dict
        GP parameters for this instance.
    X : np.ndarray
        The time coordinates for this GP instance.

    """
    kernel = GranulationKernel(params['scale'], params['period'])
    return GaussianProcess(
        kernel,
        X,
        diag=0,
        mean=params['mean']
    )


def save_cast_coverage(coverage: np.ndarray):
    """
    Safely cast coverage array so that each value is on [0,1]

    Parameters
    ----------
    coverage : np.ndarray
        The surface coverage.
    """
    coverage[coverage < 0] = 0
    coverage[coverage > 1] = 1
    return coverage


class Granulation:
    """
    Class to contain the granulation model

    Parameters
    ----------
    mean_coverage : float
        The mean surface coverage of the low-Teff regions.
    amplitude : float
        The amplitude of the low-Teff coverage oscillations.
    period : astropy.units.Quantity
        The period of low-Teff coverage oscillations.
    dteff : astropy.units.Quantity
        The temperature difference between the nominal photosphere and the low-Teff region.

    Attributes
    ----------
    params : dict
        Parameters for the gaussian process
    dteff : astropy.units.Quantity
        The temperature difference between the nominal photosphere and the low-Teff region.
    """

    def __init__(
            self,
            mean_coverage: float,
            amplitude: float,
            period: u.Quantity,
            dteff: u.Quantity,
            seed:int = np.random.randint(1, None)
    ):
        self.params = dict(
            mean=mean_coverage,
            scale=amplitude,
            period=period.to_value(time_unit)
        )
        self.dteff = dteff
        self.seed = seed
    @classmethod
    def off(
        cls,
        seed:int = 0
    ):
        """
        An instance that does not generate any granulation
        
        Parameters
        ----------
        seed : int
            Seed for the random number generator.
        """
        return cls(
            mean_coverage=0,
            amplitude=0,
            period=0*u.hr,
            dteff=0*u.K,
            seed=seed
        )
    
    def _build_gp(self, X: np.ndarray) -> GaussianProcess:
        return build_gp(self.params, X)

    def get_coverage(self, time: u.Quantity) -> np.ndarray:
        """
        Get low-Teff region coverage at each point in time.

        Parameters
        ----------
        time : astropy.units.Quantity
            The time axis to generate coverages of.

        Returns
        -------
        np.ndarray
            The coverage corresponding to each point in `time`.
        """
        key = jax.random.PRNGKey(seed=self.seed)
        gp = self._build_gp(time.to_value(time_unit))
        coverage = gp.sample(key)
        return save_cast_coverage(np.array(coverage))



"""
Flares
------

This code governs the behavior of flares.
"""
from typing import List
import typing as Typing

import numpy as np
from astropy import units as u, constants as const
from astropy.units.quantity import Quantity

from .xoflares import _flareintegralnp as flareintegral, get_light_curvenp




class StellarFlare:
    """
    Class to store and control stellar flare information

    Parameters
    ----------
    fwhm : astropy.units.Quantity
        Full-width-half-maximum of the flare
    energy : astropy.units.Quantity
        Time-integrated bolometric energy
    lat : astropy.units.Quantity
        Latitude of flare on star
    lon : astropy.units.Quantity
        Longitude of flare on star
    Teff : astropy.units.Quantity
        Blackbody temperature
    tpeak : astropy.units.Quantity
        Time of the flare peak

    Attributes
    ----------
    fwhm : astropy.units.Quantity
        Full-width-half-maximum of the flare
    energy : astropy.units.Quantity
        Time-integrated bolometric energy
    lat : astropy.units.Quantity
        Latitude of flare on star
    lon : astropy.units.Quantity
        Longitude of flare on star
    Teff : astropy.units.Quantity
        Blackbody temperature
    tpeak : astropy.units.Quantity
        Time of the flare peak

    Notes
    -----
    Flare lightcurve profiles are calculated using the `xoflares` package [1]_.
    In order to keep the multiwavenlength model self-consistent, we
    model a flare as a constant temperature surface that emits blackbody
    radiation. The lightcurve is created as the surface area grows, then decays.

    References
    ----------
    .. [1] https://github.com/mrtommyb/xoflares

    """

    def __init__(self, fwhm: Quantity, energy: Quantity, lat: Quantity, lon: Quantity, Teff: Quantity, tpeak: Quantity):
        self.fwhm = fwhm
        self.energy = energy
        self.lat = lat
        self.lon = lon
        self.Teff = Teff
        self.tpeak = tpeak

    def calc_peak_area(self) -> u.Quantity:
        """
        Calculate the flare area at its peak

        Returns
        -------
        astropy.units.Quantity
            Peak flare area
        """
        # pylint: disable-next=no-member
        time_area = self.energy/const.sigma_sb/(self.Teff**4)
        area_std = 1*u.km**2
        time_area_std = flareintegral(self.fwhm, area_std)
        area = time_area/time_area_std * area_std
        return area.to(u.km**2)

    def areacurve(self, time: Quantity):
        """
        Compute the flare area as a function of time

        Parameters
        ----------
        time : astropy.units.Quantity 
            The times at which to sample the area

        Returns
        -------
        astropy.units.Quantity 
            Area at each time

        """
        t_unit = u.day  # this is the unit of xoflares
        a_unit = u.km**2
        peak_area = self.calc_peak_area()
        areas = get_light_curvenp(
            time.to_value(t_unit),
            [self.tpeak.to_value(t_unit)],
            [self.fwhm.to_value(t_unit)],
            [peak_area.to_value(a_unit)]
        )
        return areas * a_unit

    def get_timearea(self, time: Quantity):
        """
        Calcualte the integrated time*area of the flare.

        Parameters
        ----------
        time : astropy.units.Quantity
            the times at which to sample the area.

        Returns
        -------
        astropy.units.Quantity
            The integrated time-area of the flare.
        """
        areas = self.areacurve(time)
        timearea = np.trapz(areas, time)
        return timearea.to(u.Unit('hr km2'))


class FlareGenerator:
    """
    Generator for flare events and their characteristics.

    Parameters
    ----------
    dist_teff_mean : astropy.units.Quantity
        The mean of the temperature distribution.
    dist_teff_sigma : astropy.units.Quantity
        The standard deviation of the temperature distribution.
    dist_fwhm_mean : astropy.units.Quantity
        The mean of the FWHM distribution.
    dist_fwhm_logsigma : float
        The standard deviation of the logorithm of the FWHM distribution in dex.
    alpha : float, default=-0.829
        The slope of the log frequency - log energy relationship.
    beta : float, default=26.87
        The y-intercept of the log frequency - log energy relationship.
    min_energy : astropy.units.Quantity, default=1e33 erg
        The minimum energy to consider. Set to ``np.inf*u.erg`` to disable flares.
    cluster_size : int, default=2
        The typical size of flare clusters.
    rng : numpy.random.Generator, default=numpy.random.default_rng()
        The random number generator instance to use.

    Attributes
    ----------
    energy_unit
    time_unit
    temperature_unit
    dist_teff_mean : astropy.units.Quantity
        The mean of the temperature distribution.
    dist_teff_sigma : astropy.units.Quantity
        The standard deviation of the temperature distribution.
    dist_fwhm_mean : astropy.units.Quantity
        The mean of the FWHM distribution.
    dist_fwhm_logsigma : float
        The standard deviation of the logorithm of the FWHM distribution in dex.
    alpha : float
        The slope of the log frequency - log energy relationship.
    beta : float
        The y-intercept of the log frequency - log energy relationship.
    min_energy : astropy.units.Quantity
        The minimum energy to consider. Set to ``np.inf*u.erg`` to disable flares.
    cluster_size : int
        The typical size of flare clusters.
    rng : numpy.random.Generator
        The random number generator instance to use.

    Notes
    -----
    Relationship between flare frequency is defined as

    .. math::
        \\log{(f(E \\ge E_0)~\\text{[day]})} = \\beta + \\alpha \\log{(E_0/\\text{[erg]})}

    by :cite:t:`2022AJ....164..213G`. Their study of TESS flare rates fits :math:`\\alpha = -0.829`
    and :math:`\\beta = 26.87`, which we use as the default values. This relationship is valid for
    :math:`E > 10^{33}` erg, which we use as the default minimum considered energy.

    To disable flares, set ``min_energy`` to infinity so that the expected number of flares is 0.

    """
    energy_unit = u.erg
    """
    The energy unit to standardize coversions to floats and logs.

    :type: astropy.units.Quantity
    """
    time_unit = u.day
    """
    The time unit to standardize coversions to floats and logs.

    :type: astropy.units.Quantity
    """
    temperature_unit = u.K
    """
    The temperature unit to standardize coversions to floats and logs.

    :type: astropy.units.Quantity
    """

    def __init__(
        self,
        dist_teff_mean: u.Quantity,
        dist_teff_sigma: u.Quantity,
        dist_fwhm_mean: u.Quantity,
        dist_fwhm_logsigma: float,
        alpha: float = -0.829,
        beta: float = 26.87,
        min_energy: u.Quantity = 1e33*u.erg,
        cluster_size: int = 2,
        rng: np.random.Generator = np.random.default_rng()
    ):
        self.dist_teff_mean = dist_teff_mean
        self.dist_teff_sigma = dist_teff_sigma
        self.dist_fwhm_mean = dist_fwhm_mean
        self.dist_fwhm_logsigma = dist_fwhm_logsigma
        self.min_energy = min_energy
        self.cluster_size = cluster_size
        self.alpha = alpha
        self.beta = beta
        self.rng = rng
    @classmethod
    def off(
        cls,
        rng: np.random.Generator = np.random.default_rng()
    ):
        """
        An instance that does not generate anything.
        
        Parameters
        ----------
        rng : numpy.random.Generator
            The random number generator to use.
        """
        return cls(
            dist_teff_mean=0*u.K,
            dist_teff_sigma=0*u.K,
            dist_fwhm_mean=0*u.erg,
            dist_fwhm_logsigma=0,
            alpha=-0.829,
            beta=26.87,
            min_energy=np.inf*u.erg,
            cluster_size=2,
            rng=rng
        )

    def frequency_greater_than(self, energy: u.Quantity):
        """
        The frequency of flares greater than an energy.

        Parameters
        ----------
        energy : astropy.units.Quantity
            The energy to compute the frequency of.

        Returns
        -------
        astropy.units.Quantity
            The frequency of flares with energy greater than `energy`.
        """
        return 10**self.beta * (energy.to_value(self.energy_unit))**self.alpha / self.time_unit

    def get_nexp_greater_than(self, energy: u.Quantity, time: u.Quantity) -> float:
        """
        Get the expected number of flares with energy greater than `energy` in a
        time period `time`

        Parameters
        ----------
        energy : astropy.units.Quantity
            The energy to compute the frequency of.
        time : astropy.units.Quantity
            The time duration.

        Returns
        -------
        float
            The number of flares with energy greater than `energy`.
        """
        return (self.frequency_greater_than(energy)*time).to_value(u.dimensionless_unscaled)

    def get_ntotal(self, time) -> float:
        """
        Get the total number of flares in a time period.

        Parameters
        ----------
        time : astropy.units.Quantity
            The time duration.

        Returns
        -------
        float
            The number of flares.
        """
        return self.get_nexp_greater_than(self.min_energy, time)

    def pdf(self, energy: u.Quantity):
        """
        The probability density function of flare energies.

        Parameters
        ----------
        energy : astropy.units.Quantity
            The energies at which to sample the pdf.

        Returns
        -------
        pdf : float or np.ndarray
            The probability density of energy
        """
        return (-self.alpha * energy**(self.alpha-1)/self.min_energy**self.alpha).to_value(u.dimensionless_unscaled)

    def cdf(self, energy: u.Quantity):
        """
        Get the cumulative density function.

        Parameters
        ----------
        energy : astropy.units.Quantity
            The energies at which to sample the cdf.

        Returns
        -------
        cdf : float or np.ndarray
            The cumulative density of energy
        """
        return 1 - (energy/self.min_energy).to_value(u.dimensionless_unscaled)**self.alpha

    def quantile_func(self, X: np.ndarray):
        """
        The quantile function of this flare energy distribution.

        This function maps a variable on the range [0,1) to the distribution
        defined by :math:`\\beta` and :math:`\\alpha`.

        Parameters
        ----------
        X : numpy.ndarray
            A random variable with domain [0,1)

        Returns
        -------
        energy : astropy.Units.Quantity
            The energies represented by ``X``.
        """
        X = np.atleast_1d(X)
        if not (np.all(X >= 0.) and np.all(X < 1.)):
            raise ValueError('`X` must be in [0,1)')
        return self.min_energy * (1-X)**(1/self.alpha)

    def get_peaks(self, n_flares)->np.ndarray:
        """
        Generate the times at which each flare reaches its peak.

        Parameters
        ----------
        n_flares : int
            The number of peak times to generate.

        Returns
        -------
        tpeaks : np.ndarray
            The times that the flares reach their peak in units of the total observation time.
        """
        cluster_sizes = self.rng.poisson(lam=self.cluster_size, size=n_flares)
        cluster_sizes = cluster_sizes[cluster_sizes > 0]
        cluster_sizes = cluster_sizes[np.cumsum(cluster_sizes) < n_flares]
        n_leftover = n_flares - np.sum(cluster_sizes)
        cluster_sizes = np.append(cluster_sizes, [1]*n_leftover)

        n_clusters = len(cluster_sizes)
        
        if np.sum(cluster_sizes) != n_flares:
            raise ValueError(f'Expected {n_flares} flares, got {np.sum(cluster_sizes)}')

        cluster_timescale = 1/n_clusters

        dtimes = []
        for size in cluster_sizes:
            dtimes.append(self.rng.exponential(cluster_timescale))
            flare_timescale = cluster_timescale/size/2
            for _ in range(size-1):
                dtimes.append(self.rng.exponential(flare_timescale))
        dtimes = np.array(dtimes)
        tpeaks = np.cumsum(dtimes)
        return tpeaks

    def gen_fwhm(self, n_flares: int):
        """
        Generate a FWHM set in a lognormal distribution.

        Parameters
        ----------
        n_flares : int
            The number of random FWHM lengths to generate.

        Returns
        -------
        fwhm : astropy.units.Quantity
            The FWHMs of the flares
        """
        dist_logmean_fwhm = np.log10(
            self.dist_fwhm_mean.to_value(self.time_unit))
        log_fwhm = self.rng.normal(
            loc=dist_logmean_fwhm,
            scale=self.dist_fwhm_logsigma,
            size=n_flares
        )
        fwhm = 10**log_fwhm * self.time_unit
        return fwhm

    def gen_teffs(self, n_flares):
        """
        Generate a random set of flare temperatures.

        Parameters
        ----------
        n_flares : int
            The number of random temperatures to generate.

        Returns
        -------
        teffs : astropy.units.Quantity
            The temperatures of the flares
        """

        teffs = self.rng.normal(
            loc=self.dist_teff_mean.to_value(self.temperature_unit),
            scale=self.dist_teff_sigma.to_value(self.temperature_unit),
            size=n_flares
        )*self.temperature_unit
        return teffs

    def gen_coords(self, n_flares):
        """
        Generate a set of random coordinates.

        Parameters
        ----------
        n_flares : int
            The number of random temperatures to generate.

        Returns
        -------
        lats : astropy.units.Quantity
            The stellar latitudes of the flares.
        lons : astropy.units.Quantity
            The stellar longitudes of the flares.

        Notes
        -----
        ``lats`` is generated using a quantile function for latitude.

        .. math::
            \\text{lat} = \\arcsin(2X - 1)

        For a random variable :math:`X` on [0,1).
        """
        lons = self.rng.random(size=n_flares)*360*u.deg
        lats = np.arcsin(2*self.rng.random(size=n_flares) - 1) / \
            np.pi * 180*u.deg
        return lats, lons

    def generate_flare_series(self, time: u.Quantity):
        """
        Generate a series of flares based on the distribution parameters of this
        generator.

        Parameters
        ----------
        time : astropy.units.Quantity
            The time over which the flares are observed.

        Returns
        -------
        flares : list of StellarFlare
            The flares that occur over ``time``.
        """
        nexp = self.get_ntotal(time)
        n_flares = self.rng.poisson(nexp)
        if n_flares == 0:
            return []

        energies = self.quantile_func(self.rng.random(n_flares))
        tpeaks = self.get_peaks(n_flares)*time
        fwhms = self.gen_fwhm(n_flares)
        teffs = self.gen_teffs(n_flares)
        lats, lons = self.gen_coords(n_flares)

        flares = [
            StellarFlare(
                fwhm=f,
                energy=e,
                lat=lat,
                lon=lon,
                Teff=teff,
                tpeak=t
            ) for f, e, lat, lon, teff, t in zip(fwhms, energies, lats, lons, teffs, tpeaks)
        ]
        return flares


class FlareCollection:
    """
    This class stores a series of flares and does math to turn them into lightcurves.

    Parameters
    ----------
    flares : list of `StellarFlare` or `StellarFlare`
        List of flares.

    Attributes
    ----------
    flares : list of `StellarFlare`
        List of flares.
    peaks : numpy.ndarray
        Peak times for the flares as a numpy array.
    fwhms : numpy.ndarray
        Full width at half maximum for the flares as a numpy array.

    """

    def __init__(self, flares: Typing.Union[List[StellarFlare], StellarFlare]):
        if isinstance(flares, StellarFlare):
            self.flares = [flares]
        else:
            self.flares = flares
        self.index()

    def index(self):
        """
        Get peak times and fwhms for flares as arrays.

        Returns
        -------
        None

        """
        tpeak = []
        fwhm = []
        unit = u.hr
        for flare in self.flares:
            tpeak.append(flare.tpeak.to_value(unit))
            fwhm.append(flare.fwhm.to_value(unit))
        tpeak = np.array(tpeak)*unit
        fwhm = np.array(fwhm)*unit
        self.peaks = tpeak
        self.fwhms = fwhm

    def mask(self, tstart: Quantity, tfinish: Quantity):
        """
        Create a boolean mask to indicate which flares are visible within a certain time period.

        Parameters
        ----------
        tstart : astropy.units.Quantity 
            Starting time.
        tfinish : astropy.units.Quantity 
            Ending time.

        Returns
        -------
        np.array
            Boolean array of visible flares.

        """
        padding_after = 10  # number of fwhm ouside this range a flare peak can be to still be included
        padding_before = 20
        after_start = self.peaks + padding_before*self.fwhms > tstart
        before_end = self.peaks - padding_after*self.fwhms < tfinish

        return after_start & before_end

    def get_flares_in_timeperiod(self, tstart: Quantity, tfinish: Quantity) -> List[StellarFlare]:
        """
        Generate a mask and select flares without casting to `np.ndarray`
        (use a list comprehension instead).

        Parameters
        ----------
        tstart : astropy.units.Quantity 
            Starting time.
        tfinish : astropy.units.Quantity 
            Ending time.


        Returns
        -------
        list of `StellarFlare`
            Flares that occur over specified time period.
        """
        mask = self.mask(tstart, tfinish)
        masked_flares = [flare for flare, include in zip(
            self.flares, mask) if include]
        # essentially the same as self.flares[mask], but without casting to ndarray
        return masked_flares

    def get_visible_flares_in_timeperiod(
            self,
            tstart: Quantity,
            tfinish: Quantity,
            sub_obs_coords:dict=None
        ) -> List[StellarFlare]:
        """
        Get visible flares in a given time period on a given hemisphere.

        Parameters
        ----------
        tstart : astropy.units.Quantity 
            Starting time.
        tfinish : astropy.units.Quantity 
            Ending time.
        sub_obs_coords : dict, Optional
            Coordinates defining the hemisphere. Default is None.

        Returns
        -------
        list of `StellarFlare`
            A list of flares that occur and are visible to the observer.

        """
        if sub_obs_coords is None:
            sub_obs_coords={'lat': 0*u.deg, 'lon': 0*u.deg}
        masked_flares = self.get_flares_in_timeperiod(tstart, tfinish)
        visible_flares = []
        for flare in masked_flares:
            cos_c = (np.sin(sub_obs_coords['lat']) * np.sin(flare.lat)
                     + np.cos(sub_obs_coords['lat']) * np.cos(flare.lat)
                     * np.cos(sub_obs_coords['lon']-flare.lon))
            if cos_c > 0:  # proxy for angular radius that has low computation time
                visible_flares.append(flare)
        return visible_flares

    def get_flare_integral_in_timeperiod(
        self,
        tstart: Quantity,
        tfinish: Quantity,
        sub_obs_coords:dict=None
    ):
        """
        Calculate the integrated time-area for each visible
        flare in a timeperiod.

        Parameters
        ----------
        tstart : astropy.units.Quantity 
            Starting time.
        tfinish : astropy.units.Quantity 
            Ending time.
        sub_obs_coords : dict, optional
            Coordinates defining the hemisphere. Defaults to None.

        Returns
        -------
        flare_timeareas : list of dict
            List of dictionaries containing flare temperatures and integrated
            time-areas. In the format [{'Teff':9000*u.K,'timearea'=3000*u.Unit('km2 hr)},...]
        """
        if sub_obs_coords is None:
            sub_obs_coords={'lat': 0*u.deg, 'lon': 0*u.deg}
        visible_flares = self.get_visible_flares_in_timeperiod(
            tstart, tfinish, sub_obs_coords)
        flare_timeareas = []
        time_resolution = 10*u.min
        N_steps = int(((tfinish-tstart)/time_resolution).to(u.Unit('')).value)
        time = np.linspace(tstart, tfinish, N_steps)
        for flare in visible_flares:
            timearea = flare.get_timearea(time)
            flare_timeareas.append(dict(Teff=flare.Teff, timearea=timearea))
        return flare_timeareas



"""
Coordinate Grid class
"""
from typing import Tuple, Union
import warnings

import numpy as np
from astropy import units as u


def get_lat_points(n_points: int) -> u.Quantity:
    """
    Get a 1D array of latitudes.

    Parameters
    ----------
    n_points : int
        The number of points.

    Returns
    -------
    lats : astropy.units.Quantity
        Array of latitudes.
    """
    lats = np.linspace(-90, 90, n_points)*u.deg
    return lats


def get_lon_points(n_points: int) -> u.Quantity:
    """
    Get a 1D array of longitudes.

    Parameters
    ----------
    n_points : int
        The number of points.

    Returns
    -------
    lons : astropy.units.Quantity
        Array of longitudes.
    """
    lons = np.linspace(0, 360, n_points, endpoint=False)*u.deg
    return lons


class CoordinateGrid:
    """
    Base class for all coordinate grids.
    """

    def _grid(self) -> Tuple[u.Quantity, u.Quantity]:
        raise NotImplementedError(
            'Attempted to call abstract method _grid() from base class')
    @staticmethod
    def new(grid_params: Union[int,Tuple[int, int]]):
        """
        Create an instance of a coordinate grid subclass
        from a set of parameters.
        
        Parameters
        ----------
        grid_params : int or tuple
            If a tuple, the first element is the number of latitude
            points and the second is the number of longitude points.
            If an int, it is the total number of points.
        
        Returns
        -------
        grid : SpiralGrid or RectangularGrid
            An instance of a coordinate grid subclass. If grid_params
            is a tuple, the ``grid`` is a ``RectangularGrid``. If
            ``grid_params`` is an int, the ``grid`` is a ``SpiralGrid``.
        
        Raises
        ------
        TypeError
            If grid_params is not an int or a tuple.
        """
        if isinstance(grid_params, int):
            return SpiralGrid(grid_params)
        elif isinstance(grid_params, tuple):
            return RectangularGrid(*grid_params)
        else:
            raise TypeError('grid_params must be of type int or tuple')
    
    def grid(self) -> Tuple[u.Quantity, u.Quantity]:
        """
        Get a grid of latitudes and longitudes.

        Returns
        -------
        lat : astropy.units.Quantity
            Array of latitudes.
        lon : astropy.units.Quantity
            Array of longitudes.
        """
        return self._grid()

    def cos_angle_from_disk_center(
        self,
        lat0: u.Quantity,
        lon0: u.Quantity
    ) -> np.ndarray:
        """
        Get the cosine of the angle from disk center.

        Parameters
        ----------
        lat0 : astropy.units.Quantity
            The sub-observer latitude.
        lon0 : astropy.units.Quantity
            The sub-observer longitude

        Returns
        -------
        np.ndarray
            An array of cos(x) where x is
            the angle from disk center.

        Notes
        -----
        Recall
        
        .. math::

            \\mu = \\cos{x}

        Where :math:`x` is the angle from center of the disk.
        """
        latgrid, longrid = self.grid()
        mu: u.Quantity = (np.sin(lat0) * np.sin(latgrid)
                          + np.cos(lat0) * np.cos(latgrid)
                          * np.cos(lon0-longrid))
        return mu.to_value(u.dimensionless_unscaled)

    @property
    def _area(self):
        raise NotImplementedError(
            'Attempted to call abstract method _area() from base class')

    @property
    def area(self) -> np.ndarray:
        """
        Get the area of each point as a fraction of the unit sphere.

        Returns
        -------
        np.ndarray
            The area of each point.

        """
        return self._area

    def _zeros(self, dtype='float32'):
        raise NotImplementedError(
            'Attempted to call abstract method _zeros() from base class')

    def zeros(self, dtype: str = 'float32') -> np.ndarray:
        """
        Get an array of zeros.

        Parameters
        ----------
        dtype : str, default='float32'
            Data type to pass to np.zeros.

        Returns
        -------
        np.ndarray
            An array of zeros.

        """
        return self._zeros(dtype=dtype)

    def __eq__(self, other):
        raise NotImplementedError(
            'Attempted to call abstract method __eq__() from base class')

    def _display_grid(self, nlat: int, nlon: int, data: np.ndarray):
        raise NotImplementedError(
            'Attempted to call abstract method _display_grid() from base class')

    def display_grid(
        self,
        nlat: int,
        nlon: int,
        data: Union[np.ndarray, u.Quantity]
    ) -> Tuple[np.ndarray, np.ndarray, u.Quantity]:
        """
        Resample the data to a rectangular grid.

        Parameters
        ----------
        nlat : int
            Number of latitude points to resample to.
        nlon : int
            Number of longitude points to resample to.
        data : np.ndarray or astropy.units.Quantity
            The data to display.

        Returns
        -------
        lat : np.ndarray, shape=(nlat,)
            Array of latitudes.
        lon : np.ndarray, shape=(nlon,)
            Array of longitudes.
        dat : astropy.units.Quantity, shape=(nlat,nlon)
            The resampled data.

        """
        data = data*u.dimensionless_unscaled
        unit: u.Unit = data.unit
        lat, lon, dat = self._display_grid(
            nlat, nlon, data=data.to_value(unit))
        return lat, lon, dat*unit


class RectangularGrid(CoordinateGrid):
    """
    Class to standardize the creation of latitude and longitude grids.

    This class provides a convenient way to create latitude
    and longitude grids of specified dimensions. It allows
    the creation of both one-dimensional arrays and two-dimensional
    grids of latitude and longitude points.

    Parameters
    ----------
    Nlat : int, optional (default=500)
        Number of latitude points.
    Nlon : int, optional (default=1000)
        Number of longitude points.

    Raises
    ------
    TypeError
        If Nlat or Nlon is not an integer.


    Attributes
    ----------
    Nlat : int
        Number of latitude points.
    Nlon : int
        Number of longitude points.

    Examples
    --------
    >>> grid = CoordinateGrid(Nlat=100, Nlon=200)
    >>> lats, lons = grid.oned()
    >>> print(lats.shape, lons.shape)
    (100,) (200,)
    >>> grid_arr = grid.grid()
    >>> print(grid_arr.shape)
    (100, 200)
    >>> zeros_arr = grid.zeros()
    >>> print(zeros_arr.shape)
    (200, 100)
    >>> other_grid = CoordinateGrid(Nlat=100, Nlon=200)
    >>> print(grid == other_grid)
    True

    """

    def __init__(self, nlat=500, nlon=1000):
        if not isinstance(nlat, int):
            raise TypeError('Nlat must be int')
        if not isinstance(nlon, int):
            raise TypeError('Nlon must be int')
        self.nlat = nlat
        self.nlon = nlon

    def oned(self):
        """
        Create one dimensional arrays of latitude and longitude points.

        Returns
        -------
        lats : astropy.units.Quantity , shape=(Nlat,)
            Array of latitude points.
        lons : astropy.units.Quantity , shape=(Nlon,)
            Array of longitude points.

        """
        lats = get_lat_points(self.nlat)
        lons = get_lon_points(self.nlon)
        return lats, lons

    def _grid(self):
        """
        Create a 2 dimensional grid of latitude and longitude points.

        Returns
        -------
        lats : astropy.units.Quantity , shape=(Nlat,Nlon)
            Array of latitude points.
        lons : astropy.units.Quantity , shape=(Nlat,Nlon)
            Array of longitude points.

        """
        lats, lons = self.oned()
        return np.meshgrid(lats, lons)

    def _zeros(self, dtype='float32'):
        """
        Get a grid of zeros.

        Parameters
        ----------
        dtype : str, default='float32'
            Data type to pass to np.zeros.

        Returns
        -------
        arr : np.ndarray, shape=(Nlon, Nlat)
            Grid of zeros.

        """
        return np.zeros(shape=(self.nlon, self.nlat), dtype=dtype)

    def __eq__(self, other):
        """
        Check to see if two CoordinateGrid objects are equal.

        Parameters
        ----------
        other : CoordinateGrid
            Another CoordinateGrid object.

        Returns
        -------
        bool
            Whether the two objects have equal properties.

        Raises
        ------
        TypeError
            If `other` is not a CoordinateGrid object.

        """
        if not isinstance(other, RectangularGrid):
            raise TypeError('other must be of type RectangularGrid')
        else:
            return (self.nlat == other.nlat) & (self.nlon == other.nlon)

    @property
    def _area(self) -> np.ndarray:
        """
        Get the area of each point.

        Returns
        -------
        np.ndarray
            The area of each point.

        """
        latgrid, _ = self.grid()
        jacobian: u.Quantity = np.sin(latgrid + 90*u.deg)
        norm: u.Quantity = np.sum(jacobian)
        return (jacobian/norm).to_value(u.dimensionless_unscaled)

    @property
    def dlat(self):
        """
        The latitude spacing.

        Returns
        -------
        dlat : astropy.units.Quantity
            The latitude spacing.
        """
        return 180*u.deg/(self.nlat-1)

    @property
    def dlon(self):
        """
        The longitude spacing.

        Returns
        -------
        dlon : astropy.units.Quantity
            The longitude spacing.
        """
        return 360*u.deg/(self.nlon)

    def _display_grid(
        self,
        nlat: int,
        nlon: int,
        data: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Check to make sure the grid is correct. Otherwise
        there is little to do as the grid is already rectangular.
        """
        # checks
        if not data.ndim == 2:
            raise TypeError('data must be 2 dimensional')
        if not data.shape[0] == nlon:
            raise TypeError('data must have shape (nlon, nlat)')
        if not data.shape[1] == nlat:
            raise TypeError('data must have shape (nlon, nlat)')
        lats, lons = self.oned()
        return lats.to_value(u.rad), lons.to_value(u.rad), data


class SpiralGrid(CoordinateGrid):
    """
    A class to generate a grid of points
    using a Fibonacci spiral.
    """
    GOLDEN_RATIO = 0.5*(1+np.sqrt(5))

    def __init__(self, n_points: int):
        self.n_points = n_points

    def _grid(self):
        """
        Produce a 1D grid of latitudes and longitudes.

        Returns
        -------
        lat : astropy.units.Quantity
            Array of latitudes.
        lon : astropy.units.Quantity
            Array of longitudes.

        Notes
        -----
        Algorithm from:
            https://extremelearning.com.au/how-to-evenly-distribute-points-on-a-sphere-more-effectively-than-the-canonical-fibonacci-lattice/
        """
        i = np.arange(self.n_points)
        lon = 2*np.pi*i/self.GOLDEN_RATIO
        lon = (lon % (2*np.pi))*u.rad
        colat = np.arccos(1-2*(i+0.5)/self.n_points)
        lat = (np.pi/2 - colat)*u.rad
        return lat, lon

    def _zeros(self, dtype='float32') -> np.ndarray:
        return np.zeros(shape=(self.n_points,), dtype=dtype)

    @property
    def _area(self) -> np.ndarray:
        """
        The area as a fraction of the unit sphere.
        """
        return self.zeros() + 1/self.n_points

    def __eq__(self, other) -> bool:
        if isinstance(other, CoordinateGrid):
            if isinstance(other, SpiralGrid):
                return self.n_points == other.n_points
            else:
                warnings.warn(
                    'Comparing two different subclasses of CoordinateGrid. Evaluating to False.')
                return False
        else:
            raise TypeError('other must be of type CoordinateGrid')

    def _display_grid(
        self,
        nlat: int,
        nlon: int,
        data: np.ndarray
    ):
        """
        Resample to a rectangular grid using a gaussian.
        """
        # checks
        if not data.ndim == 1:
            raise ValueError('data must be a 1D array')
        if not data.shape[0] == self.n_points:
            raise ValueError('data must have length n_points')
        data_lat, data_lon = self.grid()
        rect = RectangularGrid(nlat, nlon)

        numerator = np.zeros((nlon, nlat))
        denominator = np.zeros((nlon, nlat))
        # number density in points per steradian
        num_density = self.n_points/(4*np.pi)
        characteristic_len = 1/np.sqrt(num_density)  # characteristic length

        for value, lat, lon in zip(data, data_lat, data_lon):
            cos_r = rect.cos_angle_from_disk_center(lat, lon)
            r = np.arccos(cos_r)
            weight = np.exp(-(r/characteristic_len)**2)
            numerator += weight*value
            denominator += weight
        resampled_data = numerator/denominator
        lats, lons = rect.oned()
        return lats.to_value(u.rad), lons.to_value(u.rad), resampled_data



import pytest
import astropy.units as u
import numpy as np

from vspec_vsm.star import Star
from vspec_vsm.spots import SpotCollection, SpotGenerator
from vspec_vsm.faculae import FaculaCollection, FaculaGenerator
from vspec_vsm.flares import FlareGenerator

from vspec_vsm.granules import Granulation
from vspec_vsm.coordinate_grid import CoordinateGrid
from vspec_vsm.config import MSH

@pytest.fixture
def star():
    # Create an instance of the Star class for testing
    Teff = 3000 * u.K
    radius = 0.15 * u.Rsun
    period = 10 * u.day
    spots = SpotCollection()
    faculae = FaculaCollection()
    return Star(Teff, radius, period, spots, faculae,grid_params=(500,1000))

@pytest.fixture
def star_with_spots():
    # Create an instance of the Star class for testing
    # This one has spots
    Teff = 3000 * u.K
    radius = 0.15 * u.Rsun
    period = 10 * u.day
    faculae = FaculaCollection()
    sgen = SpotGenerator(500*MSH,0.2,2700*u.K,2600*u.K)
    spots = SpotCollection(*sgen.birth_spots(10*u.day,radius))
    return Star(Teff, radius, period, spots, faculae,
                spot_generator=sgen,grid_params=1000)
@pytest.fixture
def star_with_spots_and_fac():
    # Create an instance of the Star class for testing
    # This one has spots
    Teff = 3000 * u.K
    radius = 0.15 * u.Rsun
    period = 10 * u.day
    
    sgen = SpotGenerator(500*MSH,0.2,2700*u.K,2600*u.K)
    spots = SpotCollection(*sgen.birth_spots(10*u.day,radius))
    fgen = FaculaGenerator(
            dist_r_peak=100*u.km,
            dist_r_logsigma=0.2,
            depth=100*u.km,
            dist_life_peak=6*u.hr,
            dist_life_logsigma=0.2,
            floor_teff_slope=0*u.K/u.km,
            floor_teff_min_rad=20*u.km,
            floor_teff_base_dteff=-100*u.K,
            wall_teff_slope=0*u.K/u.km,
            wall_teff_intercept=100*u.K,
            coverage=0.01,
            grid_params=(300, 600),
            gridmaker=CoordinateGrid.new((300, 600)),
            rng=np.random.default_rng()
        )
    faculae = FaculaCollection(*fgen.birth_faculae(10*u.hr,radius))
    return Star(Teff, radius, period, spots, faculae,
                spot_generator=sgen,fac_generator=fgen)


def test_star_initialization(star:Star):
    assert star.teff == 3000 * u.K
    assert star.radius == 0.15 * u.Rsun
    assert star.period == 10 * u.day
    assert isinstance(star.spots, SpotCollection)
    assert isinstance(star.faculae, FaculaCollection)
    assert isinstance(star.gridmaker, CoordinateGrid)
    assert isinstance(star.map, u.Quantity)
    assert isinstance(star.flare_generator, FlareGenerator)
    assert isinstance(star.spot_generator, SpotGenerator)
    assert isinstance(star.fac_generator, FaculaGenerator)
    assert isinstance(star.granulation,Granulation)
    assert star.u1 == 0
    assert star.u2 == 0


def test_get_pixelmap(star:Star):
    pixelmap = star.map

    assert isinstance(pixelmap, u.Quantity)
    assert pixelmap.shape == (star.gridmaker.nlon, star.gridmaker.nlat)
    assert pixelmap.unit == u.K

def test_star_with_granulation():
    Teff = 5000 * u.K
    radius = 1 * u.Rsun
    period = 10 * u.day
    spots = SpotCollection()
    faculae = FaculaCollection()
    granulation = Granulation(0.2,0.01,5*u.day,200*u.K)
    star = Star(Teff, radius, period, spots, faculae, granulation=granulation)

    assert isinstance(star.granulation, Granulation)


def test_star_with_custom_gridmaker():
    Teff = 5000 * u.K
    radius = 1 * u.Rsun
    period = 10 * u.day
    spots = SpotCollection()
    faculae = FaculaCollection()
    star = Star(Teff, radius, period, spots, faculae,grid_params=1000)

    assert isinstance(star.gridmaker, CoordinateGrid)


def test_age_method_updates_pixelmap(star_with_spots:Star):
    # Store the initial pixelmap
    initial_pixelmap = star_with_spots.map
    

    # Age the spots and faculae
    time = 4 * u.day
    star_with_spots.age(time)

    # Check if the pixelmap has been updated
    assert not np.array_equal(star_with_spots.map, initial_pixelmap)

    # Make additional assertions based on the expected behavior of the method



def test_age_method_with_zero_time(star_with_spots:Star):
    # Add some spots and faculae to the star
    initial_num_spots = len(star_with_spots.spots.spots)
    initial_num_faculae = len(star_with_spots.faculae.faculae)

    # Age the spots and faculae with zero time
    time = 0 * u.day
    star_with_spots.age(time)

    # Check if the spots and faculae remain unchanged
    assert len(star_with_spots.spots.spots) == initial_num_spots
    assert len(star_with_spots.faculae.faculae) == initial_num_faculae

def test_transit_mask():
    Teff = 3000 * u.K
    radius = 0.15 * u.Rsun
    period = 10 * u.day
    spots = SpotCollection()
    faculae = FaculaCollection()
    star = Star(Teff, radius, period, spots, faculae,grid_params=(500,1000))
    
    mask, val = star.get_transit_mask(
        lat0=0*u.deg,
        lon0=0*u.deg,
        orbit_radius=0.05*u.AU,
        radius = 1*u.R_earth,
        phase=180*u.deg,
        inclination=90*u.deg
    )
    assert val == 1
    assert np.any(mask == 0)
    assert np.max(mask) == pytest.approx(1,abs=1e-2)
    assert np.any((mask<1) & (mask>0))
    
    
    # star-sized planet
    # mask, val = star.get_transit_mask(
    #     lat0=0*u.deg,
    #     lon0=0*u.deg,
    #     orbit_radius=1*u.AU,
    #     radius = star.radius,
    #     phase=180*u.deg,
    #     inclination=90*u.deg
    # )
    # jac = star.get_jacobian().to_value(u.dimensionless_unscaled)
    # mu = star.get_mu(lat0=0*u.deg,lon0=0*u.deg).to_value(u.dimensionless_unscaled)
    # facing = mu > 0
    # area = jac*mu
    # area_facing = np.sum(area[facing])
    # area_blocked = np.sum((mask[facing])*area[facing])
    # assert area_blocked == pytest.approx(area_facing,rel=1e-6)
    
    # rp/rs = 0.5
    # x = [8,32,128]
    # y = []
    # phases = (np.linspace(0,1,20)+180)*u.deg
    # t0 = datetime.now()
    # for rs_rp in x:
    #     # rs_rp = 4
    #     lat0 = 0*u.deg
    #     lon0=0*u.deg
    #     a = 0.05*u.AU
    #     rad = star.radius/rs_rp
    #     phase = 180.*u.deg
    #     i = 90*u.deg
    #     yy = []
    #     for phase in phases:
    #         mask, val = star.get_transit_mask(
    #             lat0=lat0,
    #             lon0=lon0,
    #             orbit_radius=a,
    #             radius = rad,
    #             phase=phase,
    #             inclination=i
    #         )
    #         jac = star.get_jacobian().to_value(u.dimensionless_unscaled)
    #         mu = star.get_mu(lat0=0*u.deg,lon0=0*u.deg).to_value(u.dimensionless_unscaled)
    #         facing = mu > 0
    #         area = jac*mu
    #         area_facing = np.sum(area[facing])
    #         area_blocked = np.sum((mask[facing])*area[facing])
    #         expected_area_blocked = area_facing/rs_rp**2
    #         extra_area_blocked = area_blocked - expected_area_blocked
    #         extra_area_blocked_div_exp = extra_area_blocked/expected_area_blocked
    #         yy.append(extra_area_blocked_div_exp*100)
    #         # unity = rs_rp**2*area_blocked/area_facing
    #         # y.append(100*(unity-1))
    #     plt.scatter(phases,yy,label=f'{rs_rp}')
    # # plt.scatter(x,y)
    # plt.xlabel('phase')
    # plt.ylabel('Extra Area blocked (%)')
    # # plt.ylabel('Divergence from expected (%)')
    # # plt.xscale('log')
    # # plt.xticks(x,labels=x)
    # plt.title(datetime.now()-t0)
    # plt.savefig('transit_bad.png',facecolor='w')
    
        
def test_calc_coverage():
    Teff = 3000 * u.K
    radius = 0.15 * u.R_sun
    period = 10 * u.day
    spots = SpotCollection()
    faculae = FaculaCollection()
    star = Star(Teff, radius, period, spots, faculae,grid_params=(500,1000),
                u1=0.1)
    
    rs_rp = 10
    lat0 = 0*u.deg
    lon0=0*u.deg
    sub_obs = {'lat':lat0,'lon':lon0}
    a = 0.05*u.AU
    rad = star.radius/rs_rp
    phase = 180.*u.deg
    i = 90*u.deg

    dat_tot, dat_cov, pl_frac = star.calc_coverage(
        sub_obs_coords=sub_obs,
        granulation_fraction=0.0
    )
    assert dat_tot[star.teff] == 1.0, 'Total coverage must be 1.0'
    assert dat_cov[star.teff] == 0.0, 'Occulted coverage must be 0.0'
    assert pl_frac == 1
    
    dat_tot, dat_cov, pl_frac = star.calc_coverage(
        sub_obs_coords=sub_obs,
        granulation_fraction=0.0,
        orbit_radius=a,
        planet_radius=rad,
        phase=phase,
        inclination=i
    )
    assert dat_tot[star.teff] == 1.0, 'Total coverage must be 1.0'
    # assert dat_cov[star.Teff] == pytest.approx(1/rs_rp**2,rel=1e-3), 'Planet occultation incorrect for phase=90deg'
    assert pl_frac == 1, 'All of planet must be visible.'
    
    dat_tot, dat_cov, pl_frac = star.calc_coverage(
        sub_obs_coords=sub_obs,
        granulation_fraction=0.0,
        orbit_radius=a,
        planet_radius=rad,
        phase=phase+0.3*u.deg,
        inclination=i
    )
    assert dat_tot[star.teff] == 1.0, 'Total coverage must be 1.0'
    # assert dat_cov[star.Teff] == pytest.approx(1/rs_rp**2,rel=1e-3), 'Planet occultation incorrect for phase=90.3deg'
    assert pl_frac == 1, 'All of planet must be visible.'
    
def test_transit_coverage():
    rp_rs = np.logspace(-2.5,-1,10)
    Teff = 3000 * u.K
    radius = 0.15 * u.R_sun
    period = 10 * u.day
    spots = SpotCollection()
    faculae = FaculaCollection()
    star = Star(Teff, radius, period, spots, faculae,grid_params=1000000)
    
    lat0 = 0*u.deg
    lon0=0*u.deg
    sub_obs = {'lat':lat0,'lon':lon0}
    a = 0.05*u.AU
    phase = 180.*u.deg
    i = 90*u.deg
    
    for r in rp_rs:
        dat_tot, dat_cov, pl_frac = star.calc_coverage(
            sub_obs_coords=sub_obs,
            granulation_fraction=0.0,
            orbit_radius=a,
            planet_radius=r*star.radius,
            phase=phase,
            inclination=i
        )
        assert dat_tot[star.teff] == 1.0, 'Total coverage must be 1.0'
        assert dat_cov[star.teff] == pytest.approx(r**2,rel=5e-2), f'Planet occultation incorrect for rp/rs = {r}'
        assert pl_frac == 1, 'All of planet must be visible.'


# def test_add_fac_to_map():
#     Teff = 3000 * u.K
#     radius = 0.15 * u.Rsun
#     period = 10 * u.day
#     sgen = SpotGenerator(5000*MSH,0.2,2900*u.K,2800*u.K,coverage=0.4,init_area=500*MSH)
#     spots = SpotCollection(*sgen.birth_spots(10*u.day,radius))
#     fgen = FaculaGenerator(dist_r_peak = 0.1*u.R_sun,coverage=0.4)
#     # facula = Facula(0*u.deg,40*u.deg,0.1*u.R_sun,0.02*u.R_sun,2700*u.K,3700*u.K,1*u.hr,True,Zw=10000*u.km)
#     faculae = FaculaCollection(*fgen.birth_faculae(50*u.hr,radius,Teff))
#     # faculae = FaculaCollection(facula)
#     star = Star(Teff, radius, period, spots, faculae)
    
#     # fac:Facula = star.faculae.faculae[0]
#     # inside_fac = fac.map_pixels(radius)
#     lon0,lat0 = 0*u.deg,0*u.deg
#     # angle = get_angle_between(lat0,lon0,fac.lat,fac.lon)
#     # fracs = fac.fractional_effective_area(angle)
#     # N_fac_pix = np.sum(inside_fac)
#     # N_wall = N_fac_pix * list(fracs.values())[0]
#     # N_floor = N_fac_pix * list(fracs.values())[1]

#     # mu = star.get_mu(lat0,lon0)
#     # mu_of_fac_pix = mu[inside_fac]
#     # border_mu = np.percentile(mu_of_fac_pix,100*list(fracs.values())[1].value)
#     # wall_pix = inside_fac & (mu >= border_mu)
#     # floor_pix = inside_fac & (mu < border_mu)
#     # wall_teff,floor_teff = fracs.keys()
#     # map = star.map
#     # map[wall_pix] = wall_teff
#     # map[floor_pix] = floor_teff

   

#     lats,lons = star.gridmaker.oned()
#     proj = ccrs.Orthographic(
#                 central_longitude=0,central_latitude=0)
#     fig = plt.figure()
#     gs = fig.add_gridspec(1,1)
#     ax = fig.add_subplot(gs[0,0],projection=proj, fc="r")
#     im = ax.pcolormesh(lons.value,lats.value,star.add_faculae_to_map(lat0,lon0).T,transform=ccrs.PlateCarree(),cmap='viridis')
#     # im = ax.pcolormesh(lons.value,lats.value,map.T.value,transform=ccrs.PlateCarree(),cmap='viridis')
#     fig.colorbar(im,ax=ax)
#     fig.show()
#     0


# def test_transit_mask():
#     Teff = 3000 * u.K
#     radius = 0.15 * u.R_sun
#     period = 10 * u.day
#     sgen = SpotGenerator(1000*MSH,0.2,2900*u.K,2800*u.K,coverage=0.4,init_area=500*MSH,distribution='iso')
#     spots = SpotCollection(*sgen.birth_spots(100*u.day,radius))
#     fgen = FaculaGenerator(dist_r_peak = 0.1*u.R_sun,coverage=0.4)
#     # facula = Facula(0*u.deg,40*u.deg,0.1*u.R_sun,0.02*u.R_sun,2700*u.K,3700*u.K,1*u.hr,True,Zw=10000*u.km)
#     faculae = FaculaCollection(*fgen.birth_faculae(0*u.hr,radius,Teff))
#     # faculae = FaculaCollection(facula)
#     star = Star(Teff, radius, period, spots, faculae,Nlon=1000,Nlat=500)
#     pmap = star.map.value

#     lon0,lat0 = 0*u.deg,0*u.deg
    
#     inclination = 90.*u.deg
#     phase = 0.8*u.deg#+180*u.deg
#     planet_radius = 0.1*radius
#     semimajor_axis = 0.05*u.AU
#     lat,lon = star.gridmaker.oned()
#     # llat,llon = star.gridmaker.grid()
#     # llon = llon-180*u.deg
#     # x,y = proj_ortho(lat0,lon0,llat,llon)

#     # x_pl = (semimajor_axis/radius * np.sin(phase)).to_value(u.dimensionless_unscaled)
#     # y_pl = (semimajor_axis/radius * np.cos(phase)* np.cos(inclination)).to_value(u.dimensionless_unscaled)
#     # rad_pl = (planet_radius/radius).to_value(u.dimensionless_unscaled)
#     # rad_map = np.sqrt((x-x_pl)**2 + (y-y_pl)**2)
#     transited,pl_frac = star.get_transit_mask(lat0,lon0,semimajor_axis,planet_radius,phase,inclination)
#     visible = np.where(~transited,np.nan,0)



#     # plt.scatter(x,y,c=pmap,s=10,cmap='viridis')
#     # plt.colorbar()

#     fig = plt.figure()
#     gs = fig.add_gridspec(1,2)
#     proj = ccrs.Orthographic(
#                 central_longitude=lon0.to_value(u.deg),central_latitude=lat0.to_value(u.deg))
#     ax = fig.add_subplot(gs[0,0],projection=proj, fc="r")
#     im = ax.pcolormesh(lon.value,lat.value,pmap.T,transform=ccrs.PlateCarree(),cmap='viridis')
#     zorder = 100 if pl_frac == 1. else -100
#     ax.contourf(lon.value,lat.value,visible.T,transform=ccrs.PlateCarree(),colors='k',alpha=1,zorder=zorder)
#     fig.colorbar(im,ax=ax)
#     # ax2 = fig.add_subplot(gs[0,1])
#     # ax2.set_aspect('equal')
#     # im = ax2.scatter(x,y,c=pmap,s=5,cmap='viridis')
#     # fig.colorbar(im,ax=ax2)
#     fig.show()

# def test_calc_coverage():
#     Teff = 3000 * u.K
#     radius = 0.15 * u.R_sun
#     period = 10 * u.day
#     sgen = SpotGenerator(1000*MSH,0.2,2900*u.K,2800*u.K,coverage=0.4,init_area=500*MSH,distribution='iso')
#     spots = SpotCollection(*sgen.birth_spots(100*u.day,radius))
#     fgen = FaculaGenerator(dist_r_peak = 0.1*u.R_sun,coverage=0.4)
#     # facula = Facula(0*u.deg,40*u.deg,0.1*u.R_sun,0.02*u.R_sun,2700*u.K,3700*u.K,1*u.hr,True,Zw=10000*u.km)
#     faculae = FaculaCollection(*fgen.birth_faculae(0*u.hr,radius,Teff))
#     # faculae = FaculaCollection(facula)
#     star = Star(Teff, radius, period, spots, faculae,Nlon=1000,Nlat=500,granulation=Granulation(0,0,5*u.day,200*u.K))
    
#     inclination = 90.*u.deg
#     phase = 0.*u.deg+180*u.deg
#     planet_radius = 0.1*radius
#     semimajor_axis = 0.05*u.AU
#     lon0,lat0 = 0*u.deg,0*u.deg
#     total,covered,pl_frac = star.calc_coverage(
#         {'lat':lat0,'lon':lon0},0.0,semimajor_axis,planet_radius,phase,inclination
#     )
#     0

# def test_plot_surface():
#     Teff = 3000 * u.K
#     radius = 0.15 * u.R_sun
#     period = 10 * u.day
#     sgen = SpotGenerator(1000*MSH,0.2,2900*u.K,2800*u.K,coverage=0.4,init_area=500*MSH,distribution='iso')
#     spots = SpotCollection(*sgen.birth_spots(100*u.day,radius))
#     fgen = FaculaGenerator(dist_r_peak = 0.2*u.R_sun,coverage=0.4)
#     # facula = Facula(0*u.deg,40*u.deg,0.1*u.R_sun,0.02*u.R_sun,2700*u.K,3700*u.K,1*u.hr,True,Zw=10000*u.km)
#     faculae = FaculaCollection(*fgen.birth_faculae(10*u.hr,radius,Teff))
#     for fac in faculae.faculae:
#         fac.Zw = 0.01*u.R_sun
#     # faculae = FaculaCollection(facula)
#     star = Star(Teff, radius, period, spots, faculae,Nlon=1000,Nlat=500,u1=0.4,u2=-1)
#     lon0,lat0 = 0*u.deg,0*u.deg
#     inclination = 90.*u.deg
#     phase = -0.0*u.deg+180*u.deg
#     planet_radius = 0.1*radius
#     semimajor_axis = 0.05*u.AU
#     proj = ccrs.Orthographic(
#             central_latitude=lat0.to_value(u.deg),
#             central_longitude=lon0.to_value(u.deg)
#     )
#     fig,axes = plt.subplots(1,1,subplot_kw={'projection': proj },figsize=(5,4))
#     # for i,ax in enumerate(axes):
#     star.plot_surface(lat0,lon0,axes,semimajor_axis,planet_radius,
#                         phase,inclination)
#     0





if __name__ in '__main__':
    pytest.main(args=[__file__])


"""
Test module for VSPEC Granulation model
"""
import numpy as np
from tinygp import GaussianProcess
from astropy import units as u
from vspec_vsm import granules

def test_GranulationKernel():
    """
    Test for `granules.GranulationKernel`
    """
    scale=0.1
    period=3
    k = granules.GranulationKernel(scale,period)
    k.evaluate(0,1)

def test_build_gp():
    """
    Test for `granules.build_gp()`
    """
    params = dict(
            mean=0.5,
            scale=0.1,
            period=3
    )
    X = np.linspace(0,10,51)
    gp = granules.build_gp(params,X)
    assert isinstance(gp,GaussianProcess)

def test_safe_cast_coverage():
    """
    Test for `granules.safe_cast_coverage()`
    """
    a = np.linspace(0,1,5)
    assert np.all(a == granules.save_cast_coverage(a))
    a = np.zeros(5)+1.1
    assert np.all(granules.save_cast_coverage(a)==1)
    a = np.zeros(5)-0.1
    assert np.all(granules.save_cast_coverage(a)==0)

def test_granulation_init():
    """
    Test for `granules.Granulation.__init__()`
    """
    gran = granules.Granulation(0.2,0.01,3*u.day,200*u.K)
    assert hasattr(gran,'params')

def test_granulation_build_gp():
    gran = granules.Granulation(0.2,0.01,3*u.day,200*u.K)
    X = np.linspace(0,10,51)
    # pytest: disable-next:protected-access
    gp = gran._build_gp(X)
    assert isinstance(gp,GaussianProcess)

def test_granulation_get_coverage():
    t = np.linspace(0,10,51)*u.day
    gran = granules.Granulation(0.2,0.01,3*u.day,200*u.K)
    coverage = gran.get_coverage(t)
    assert not np.any(coverage > 1)
    assert not np.any(coverage < 0)
    
